from __future__ import annotations

from datetime import datetime, timedelta
from typing import Callable, Iterable
from uuid import uuid4

from skillos.context_monitor import ContextMonitor, ContextSignal
from skillos.suggestions import (
    SuggestionPreferences,
    SuggestionPreferencesStore,
    SuggestionRecord,
    SuggestionStore,
    _utc_now,
)


class SuggestionScheduler:
    def __init__(
        self,
        context_monitor: ContextMonitor,
        suggestion_store: SuggestionStore,
        preferences_store: SuggestionPreferencesStore,
        now_provider: Callable[[], datetime] = _utc_now,
    ) -> None:
        self._monitor = context_monitor
        self._store = suggestion_store
        self._preferences = preferences_store
        self._now = now_provider

    def run(self, signals: Iterable[ContextSignal]) -> list[SuggestionRecord]:
        preferences = self._preferences.load()
        if not preferences.opt_in:
            return []
        now = self._now()
        if preferences.snoozed_until and now < preferences.snoozed_until:
            return []
        relevant = self._monitor.find_relevant(list(signals), now)
        if not relevant:
            return []
        records = self._store.load()
        suggestions: list[SuggestionRecord] = []
        for signal in relevant:
            if not _within_limits(records, preferences, now):
                break
            suggestion = _build_suggestion(signal, now)
            records.append(suggestion)
            suggestions.append(suggestion)
        if suggestions:
            self._store.save(records)
        return suggestions


def _within_limits(
    records: list[SuggestionRecord],
    preferences: SuggestionPreferences,
    now: datetime,
) -> bool:
    if preferences.max_per_day <= 0:
        return False
    if preferences.min_interval_minutes > 0:
        last_created = _last_created_at(records)
        if last_created and now - last_created < timedelta(
            minutes=preferences.min_interval_minutes
        ):
            return False
    recent = _recent_records(records, now)
    if len(recent) >= preferences.max_per_day:
        return False
    return True


def _recent_records(
    records: list[SuggestionRecord], now: datetime
) -> list[SuggestionRecord]:
    cutoff = now - timedelta(days=1)
    return [record for record in records if record.created_at >= cutoff]


def _last_created_at(records: list[SuggestionRecord]) -> datetime | None:
    if not records:
        return None
    return max(record.created_at for record in records)


def _build_suggestion(signal: ContextSignal, now: datetime) -> SuggestionRecord:
    message = f"Upcoming {signal.source}: {signal.summary}"
    return SuggestionRecord(
        suggestion_id=uuid4().hex,
        source=signal.source,
        summary=signal.summary,
        message=message,
        created_at=now,
        status="created",
    )



from __future__ import annotations

from pathlib import Path

import yaml

from skillos.skills.errors import SkillValidationError
from skillos.skills.models import SkillMetadata


def scaffold_skill(skill_id: str, root: Path) -> tuple[Path, Path]:
    domain, name = _parse_skill_id(skill_id)
    root_path = Path(root)
    metadata_path = root_path / "metadata" / domain
    implementation_path = root_path / "implementations" / domain

    metadata_path.mkdir(parents=True, exist_ok=True)
    implementation_path.mkdir(parents=True, exist_ok=True)
    _ensure_package(root_path / "implementations")
    _ensure_package(implementation_path)

    metadata_file = metadata_path / f"{name}.yaml"
    implementation_file = implementation_path / f"{name}.py"

    if metadata_file.exists() or implementation_file.exists():
        raise FileExistsError("Skill already exists")

    entrypoint = f"implementations.{domain}.{name}:run"
    metadata = SkillMetadata(
        id=skill_id,
        name=name.replace("_", " ").title(),
        description=f"Auto-generated skill for {skill_id}",
        version="0.1.0",
        entrypoint=entrypoint,
        tags=[domain],
    )
    metadata_file.write_text(
        yaml.safe_dump(
            metadata.model_dump(exclude_defaults=True, exclude_none=True),
            sort_keys=False,
        ),
        encoding="utf-8",
    )
    implementation_file.write_text(_implementation_template(skill_id), encoding="utf-8")
    return metadata_file, implementation_file


def _parse_skill_id(skill_id: str) -> tuple[str, str]:
    if "/" not in skill_id:
        raise SkillValidationError("Skill id must be in 'domain/name' format")
    domain, name = skill_id.split("/", 1)
    if not domain or not name:
        raise SkillValidationError("Skill id must be in 'domain/name' format")
    return domain, name


def _ensure_package(path: Path) -> None:
    path.mkdir(parents=True, exist_ok=True)
    init_file = path / "__init__.py"
    if not init_file.exists():
        init_file.write_text("", encoding="utf-8")


def _implementation_template(skill_id: str) -> str:
    return (
        "def run(payload: str = \"ok\") -> str:\n"
        f"    return \"{skill_id} executed\"\n"
    )



from __future__ import annotations

from contextlib import contextmanager
from dataclasses import dataclass
import inspect
import importlib
from importlib import util as importlib_util
import os
from pathlib import Path
import sys
from typing import Iterable, Any
import asyncio

from skillos.skills.loader import load_skill_file
from skillos.skills.models import SkillMetadata


@dataclass(frozen=True)
class SkillRecord:
    metadata: SkillMetadata
    source: Path


class SkillRegistry:
    def __init__(self, root: Path) -> None:
        self._root = Path(root)
        self._skills: dict[str, SkillRecord] = {}
        self._refresh_token: tuple[int, float] | None = None
        # We no longer inject root into sys.path to avoid "Dependency Hell"
        # Modules should be loaded via standard importlib mechanisms or resolved by file path.

    @property
    def root(self) -> Path:
        return self._root

    @property
    def metadata_path(self) -> Path:
        return self._root / "metadata"

    def load_all(self, path: Path | None = None) -> dict[str, SkillRecord]:
        metadata_path = Path(path) if path else self.metadata_path
        if path:
            self._root = metadata_path.parent
                
        self._skills = {}

        # Purge modules to avoid cross-test contamination or stale imports
        self._purge_modules()

        # 1. Load from YAML metadata (Legacy/Standard way)
        if metadata_path.exists():
            for file_path in _iter_skill_files(metadata_path):
                metadata = load_skill_file(file_path)
                self._skills[metadata.id] = SkillRecord(metadata=metadata, source=file_path)

        # 2. Load from Python implementations (Zero-YAML way)
        # Scan 'implementations' folder for functions decorated with @skill
        impl_path = self._root / "implementations"
        if impl_path.exists():
            for py_file in sorted(impl_path.rglob("*.py")):
                # Calculate module name relative to implementations or root
                try:
                    rel_path = py_file.relative_to(self._root)
                    # Handle __init__.py specifically:
                    # implementations/foo/__init__.py -> implementations.foo
                    # implementations/__init__.py -> implementations
                    if py_file.name == "__init__.py":
                         module_name = ".".join(rel_path.parent.parts)
                    else:
                         module_name = ".".join(rel_path.with_suffix("").parts)
                    
                    
                    # Temporarily load module to inspect functions
                    try:
                        module = _load_module_safe(module_name, py_file)
                        
                        for attr_name in dir(module):
                            obj = getattr(module, attr_name)
                            if callable(obj) and hasattr(obj, "_skill_metadata"):
                                metadata = obj._skill_metadata
                                if metadata.id not in self._skills:
                                    self._skills[metadata.id] = SkillRecord(
                                        metadata=metadata, 
                                        source=py_file
                                    )
                    except Exception as e:
                         # Skip files that fail to load, but log it for debug
                         print(f"Failed to load skill from {py_file}: {e}", file=sys.stderr)
                         pass
                except ValueError:
                    continue

        self._refresh_token = self._compute_refresh_token()
        return self._skills

    def reload(self) -> dict[str, SkillRecord]:
        self._purge_modules()
        return self.load_all(self.metadata_path)

    def reload_if_changed(self) -> dict[str, SkillRecord] | None:
        token = self._compute_refresh_token()
        if self._refresh_token is None or token != self._refresh_token:
            return self.load_all(self.metadata_path)
        return None

    def _purge_modules(self) -> None:
        """Purge dynamic skill modules to ensure fresh reloading."""
        to_delete = []
        for m in list(sys.modules.keys()):
            if m.startswith("implementations.") or m == "implementations":
                to_delete.append(m)
        
        for m in to_delete:
            del sys.modules[m]
            
        importlib.invalidate_caches()

    def _compute_refresh_token(self) -> tuple[int, float]:
        files: list[Path] = []
        metadata_path = self.metadata_path
        if metadata_path.exists():
            files.extend(_iter_skill_files(metadata_path))
        implementations_path = self._root / "implementations"
        if implementations_path.exists():
            files.extend(sorted(implementations_path.rglob("*.py")))
        max_mtime = 0.0
        for path in files:
            try:
                mtime = path.stat().st_mtime
            except OSError:
                continue
            if mtime > max_mtime:
                max_mtime = mtime
        return (len(files), max_mtime)

    def get(self, skill_id: str) -> SkillMetadata | None:
        record = self._skills.get(skill_id)
        return record.metadata if record else None

    def search(self, query: str, limit: int = 5) -> list[SkillMetadata]:
        query_lower = query.lower()
        results: list[SkillMetadata] = []
        for record in self._skills.values():
            metadata = record.metadata
            haystack = " ".join(
                [metadata.id, metadata.name, metadata.description, *metadata.tags]
            ).lower()
            if query_lower in haystack:
                results.append(metadata)
        return results[:limit]

    def execute(self, skill_id: str, **kwargs):
        """Synchronous execution entrypoint."""
        from skillos.composition import CompositionEngine, CompositionStore
        from skillos.kernel import get_kernel

        record = self._skills.get(skill_id)
        if not record:
            raise KeyError(f"Unknown skill: {skill_id}")

        allow_inactive = bool(kwargs.pop("allow_inactive", False))
        payload = kwargs.pop("payload", "ok")
        role = kwargs.pop("role", None)
        attributes = kwargs.pop("attributes", None)
        approval_status = kwargs.pop("approval_status", None)
        approval_token = kwargs.pop("approval_token", None)
        charge_budget = kwargs.pop("charge_budget", True)
        session_context = kwargs.pop("session_context", None)
        
        # Composition engine check
        store = CompositionStore(self._root)
        
        with _skill_root_env(self._root):
            if store.exists(skill_id):
                engine = CompositionEngine(self, store)
                return engine.execute(
                    skill_id,
                    payload,
                    allow_inactive=allow_inactive,
                    role=role,
                    attributes=attributes,
                    approval_status=approval_status,
                    approval_token=approval_token,
                    charge_budget=charge_budget,
                    session_context=session_context,
                )
            
            kernel = get_kernel(record.metadata, root_path=str(self._root))
            return kernel.execute(
                record.metadata,
                payload,
                role=role,
                attributes=attributes,
                approval_status=approval_status,
                approval_token=approval_token,
                charge_budget=charge_budget,
                root=self._root,
                session_context=session_context,
                **kwargs
            )

    async def execute_async(self, skill_id: str, **kwargs):
        """Asynchronously execution entrypoint."""
        from skillos.composition import CompositionStore
        from skillos.kernel import get_kernel

        record = self._skills.get(skill_id)
        if not record:
            raise KeyError(f"Unknown skill: {skill_id}")

        store = CompositionStore(self._root)
        
        if store.exists(skill_id):
            # Fallback for composition skills (likely sync)
            return await asyncio.to_thread(self.execute, skill_id, **kwargs)

        kernel = get_kernel(record.metadata, root_path=str(self._root))
        payload = kwargs.pop("payload", "ok")
        role = kwargs.pop("role", None)
        attributes = kwargs.pop("attributes", None)
        approval_status = kwargs.pop("approval_status", None)
        approval_token = kwargs.pop("approval_token", None)
        charge_budget = kwargs.pop("charge_budget", True)
        session_context = kwargs.pop("session_context", None)

        return await kernel.execute_async(
            record.metadata,
            payload,
            role=role,
            attributes=attributes,
            approval_status=approval_status,
            approval_token=approval_token,
            charge_budget=charge_budget,
            root=self._root,
            session_context=session_context,
            **kwargs
        )



def resolve_entrypoint(entrypoint: str, root: Path | None = None) -> Any:
    """Resolve a skill entrypoint string (module:function) to a callable."""
    module_path, func_name = entrypoint.split(":", 1)
    
    # 1. Try standard import first (best practice)
    try:
        module = importlib.import_module(module_path)
        return getattr(module, func_name)
    except ImportError:
        pass
        
    # 2. Try to resolve as a file path relative to root
    if root:
        # Convert dotted path to file path segments
        # e.g. "implementations.chat" -> "implementations/chat.py"
        rel_path = module_path.replace(".", "/") + ".py"
        file_path = (Path(root) / rel_path).resolve()
        
        if file_path.exists():
            return _load_from_file(module_path, file_path, func_name)

        # Fallback: try looking in 'implementations' folder if not specified
        # e.g. "chat" -> "implementations/chat.py"
        if "implementations" not in module_path:
            alt_path = (Path(root) / "implementations" / rel_path).resolve()
            if alt_path.exists():
                return _load_from_file(module_path, alt_path, func_name)

    raise ImportError(
        f"Could not resolve skill entrypoint: {entrypoint}. "
        f"Checked standard imports and relative paths in {root}."
    )


def _load_from_file(module_name: str, file_path: Path, func_name: str) -> Any:
    """Helper to load a module from a specific file path."""
    module = _load_module_safe(module_name, file_path)
    try:
        return getattr(module, func_name)
    except AttributeError:
        raise ImportError(f"Module {module_name} has no attribute '{func_name}'")


def _load_module_safe(module_name: str, file_path: Path) -> Any:
    """
    Safely load a module from a file path, creating parent packages if needed
    and cleaning up sys.modules on failure to prevent poisoning.
    """
    # 1. Ensure parent packages exist in sys.modules to support relative imports
    parts = module_name.split(".")
    for i in range(1, len(parts)):
        parent_name = ".".join(parts[:i])
        if parent_name not in sys.modules:
            # Create a dummy namespace package for the parent
            parent_mod = importlib.util.module_from_spec(
                importlib.util.spec_from_loader(parent_name, loader=None)
            )
            # We assume the parent path is the directory above
            # This is a bit heuristical but necessary for "implementations" root
            try:
                # If we are loading implementations.subdir, parent is implementations
                # We need to find the correct path for it.
                # If parts are ['implementations', 'foo'], i=1 -> 'implementations'
                # parent path is file_path.parents[len(parts) - i] ?
                # Start from file_path directory
                # If path is implementations/foo/bar.py, valid parts are implementations.foo.bar
                # module_name = implementations.foo.bar
                # i=1: parent=implementations. path should optionally point to implementations dir
                # i=2: parent=implementations.foo. path should point to foo dir
                
                # Check if it actually maps to a directory
                depth = len(parts) - i
                if file_path.name == "__init__.py":
                    depth += 1
                    
                parent_dir = file_path.parents[depth-1] # 0 is dir containing file
                if parent_dir.exists():
                    parent_mod.__path__ = [str(parent_dir)]
            except IndexError:
                pass
                
            sys.modules[parent_name] = parent_mod

    spec = importlib_util.spec_from_file_location(module_name, file_path)
    if not spec or not spec.loader:
        raise ImportError(f"Could not create module spec from {file_path}")
    
    module = importlib_util.module_from_spec(spec)
    sys.modules[module_name] = module
    
    try:
        spec.loader.exec_module(module)
        return module
    except Exception as e:
        # CLEANUP: Remove module from sys.modules to prevent poisoning on retry/reload
        # We also attempt to cleanup parents if we just created them, but that's risky if used by others.
        # Minimal safety: remove the leaf node.
        if module_name in sys.modules:
            del sys.modules[module_name]
        raise ImportError(f"Failed to execute module {module_name} from {file_path}: {e}") from e

def _iter_skill_files(metadata_path: Path) -> Iterable[Path]:
    yaml_files = list(metadata_path.rglob("*.yaml"))
    yml_files = list(metadata_path.rglob("*.yml"))
    return sorted(yaml_files + yml_files)

# Update Zero-YAML scanning to calculate correct module name for __init__.py
# (This logic belongs in load_all, updating it there as well logic below implies modification to _load_from_file)





@contextmanager
def _skill_root_env(root: Path) -> Iterable[None]:
    previous = os.environ.get("SKILLOS_ROOT")
    os.environ["SKILLOS_ROOT"] = str(root)
    try:
        yield
    finally:
        if previous is None:
            os.environ.pop("SKILLOS_ROOT", None)
        else:
            os.environ["SKILLOS_ROOT"] = previous



from __future__ import annotations
from datetime import datetime
from uuid import uuid4
from typing import Dict, List, Optional
from pydantic import BaseModel, Field

def _utc_now() -> datetime:
    return datetime.now().astimezone()

class Message(BaseModel):
    role: str  # "user", "assistant", "system"
    content: str
    timestamp: datetime = Field(default_factory=_utc_now)
    metadata: Dict[str, object] = Field(default_factory=dict)

class Session(BaseModel):
    id: str = Field(default_factory=lambda: uuid4().hex)
    user_id: Optional[str] = None
    messages: List[Message] = Field(default_factory=list)
    context: Dict[str, object] = Field(default_factory=dict)
    created_at: datetime = Field(default_factory=_utc_now)
    updated_at: datetime = Field(default_factory=_utc_now)

    def add_message(self, role: str, content: str, metadata: Dict[str, object] | None = None) -> None:
        self.messages.append(Message(role=role, content=content, metadata=metadata or {}))
        self.updated_at = _utc_now()

    def update_context(self, key: str, value: object) -> None:
        self.context[key] = value
        self.updated_at = _utc_now()



from __future__ import annotations

from contextlib import contextmanager
from dataclasses import dataclass, field
import json
import time
from typing import Callable, Iterable


PromptFn = Callable[[str], None]


def _default_step_prompt(message: str) -> None:
    input(f"{message} (press Enter) ")


@dataclass(frozen=True)
class StepController:
    prompt: PromptFn = _default_step_prompt
    show_inputs: bool = False
    show_outputs: bool = False
    show_timing: bool = False

    def pause(
        self,
        name: str,
        *,
        inputs: dict[str, object] | None = None,
        outputs: dict[str, object] | None = None,
        duration_ms: float | None = None,
    ) -> None:
        parts = [f"step: {name}"]
        if self.show_inputs and inputs is not None:
            parts.append(
                f"inputs={json.dumps(inputs, ensure_ascii=True, default=str)}"
            )
        if self.show_outputs and outputs is not None:
            parts.append(
                f"outputs={json.dumps(outputs, ensure_ascii=True, default=str)}"
            )
        if self.show_timing and duration_ms is not None:
            parts.append(f"duration_ms={duration_ms:.2f}")
        self.prompt(" | ".join(parts))


@dataclass(frozen=True)
class DebugTraceConfig:
    capture_inputs: bool
    capture_outputs: bool
    capture_timing: bool


@dataclass(frozen=True)
class TraceStep:
    name: str
    inputs: dict[str, object] | None
    outputs: dict[str, object] | None
    duration_ms: float | None


@dataclass
class DebugTrace:
    config: DebugTraceConfig
    step_controller: StepController | None = None
    steps: list[TraceStep] = field(default_factory=list)

    def add_step(
        self,
        name: str,
        *,
        inputs: dict[str, object] | None = None,
        outputs: dict[str, object] | None = None,
        duration_ms: float | None = None,
    ) -> None:
        self.steps.append(
            TraceStep(
                name=name,
                inputs=inputs if self.config.capture_inputs else None,
                outputs=outputs if self.config.capture_outputs else None,
                duration_ms=duration_ms if self.config.capture_timing else None,
            )
        )

    @contextmanager
    def step(
        self, name: str, *, inputs: dict[str, object] | None = None
    ) -> Iterable[dict[str, object]]:
        start = time.perf_counter()
        outputs: dict[str, object] = {}
        try:
            yield outputs
        finally:
            capture_timing = self.config.capture_timing or (
                self.step_controller and self.step_controller.show_timing
            )
            duration_ms = (
                (time.perf_counter() - start) * 1000 if capture_timing else None
            )
            self.add_step(
                name,
                inputs=inputs,
                outputs=outputs,
                duration_ms=duration_ms,
            )
            if self.step_controller:
                self.step_controller.pause(
                    name,
                    inputs=inputs if self.config.capture_inputs else None,
                    outputs=outputs if self.config.capture_outputs else None,
                    duration_ms=duration_ms,
                )


@contextmanager
def trace_step(
    trace: DebugTrace | None,
    name: str,
    *,
    inputs: dict[str, object] | None = None,
) -> Iterable[dict[str, object]]:
    if trace is None:
        yield {}
        return
    with trace.step(name, inputs=inputs) as outputs:
        yield outputs


def render_trace(trace: DebugTrace) -> str:
    lines = ["trace:"]
    for step in trace.steps:
        line = f"- {step.name}"
        if step.duration_ms is not None:
            line += f" ({step.duration_ms:.2f}ms)"
        lines.append(line)
        if step.inputs is not None:
            lines.append(
                f"  inputs: {json.dumps(step.inputs, ensure_ascii=True)}"
            )
        if step.outputs is not None:
            lines.append(
                f"  outputs: {json.dumps(step.outputs, ensure_ascii=True)}"
            )
    return "\n".join(lines)



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
import json
import os
from pathlib import Path
from typing import Callable

from skillos.storage import atomic_write_text, file_lock
from skillos.tenancy import resolve_tenant_root
from skillos.storage_backend import (
    pg_connect,
    require_postgres_dsn,
    resolve_tenant_id,
    storage_backend_from_env,
)

DEFAULT_TTL_SECONDS = 300


def default_idempotency_path(root: Path) -> Path:
    root_path = resolve_tenant_root(root)
    return root_path / "runtime" / "idempotency.json"


def idempotency_ttl_from_env() -> int:
    raw = os.getenv("SKILLOS_IDEMPOTENCY_TTL_SECONDS")
    if raw is None:
        return DEFAULT_TTL_SECONDS
    try:
        return max(0, int(raw))
    except ValueError:
        return DEFAULT_TTL_SECONDS


def _utc_now() -> datetime:
    return datetime.now(timezone.utc)


def _ensure_utc(value: datetime) -> datetime:
    if value.tzinfo is None:
        return value.replace(tzinfo=timezone.utc)
    return value.astimezone(timezone.utc)


def _format_datetime(value: datetime) -> str:
    return _ensure_utc(value).isoformat()


def _parse_datetime(value: str) -> datetime:
    parsed = datetime.fromisoformat(value)
    return _ensure_utc(parsed)


@dataclass(frozen=True)
class IdempotencyEntry:
    source: str
    skill_id: str
    idempotency_key: str
    expires_at: datetime

    def to_dict(self) -> dict[str, str]:
        return {
            "source": self.source,
            "skill_id": self.skill_id,
            "idempotency_key": self.idempotency_key,
            "expires_at": _format_datetime(self.expires_at),
        }


@dataclass(frozen=True)
class IdempotencyDecision:
    allowed: bool
    expires_at: datetime | None


class IdempotencyStore:
    def __init__(
        self,
        path: Path,
        *,
        now_provider: Callable[[], datetime] = _utc_now,
    ) -> None:
        self.path = Path(path)
        self._now = now_provider

    def check_and_record(
        self,
        source: str,
        skill_id: str,
        idempotency_key: str,
        *,
        ttl_seconds: int = DEFAULT_TTL_SECONDS,
    ) -> IdempotencyDecision:
        with file_lock(self.path):
            now = _ensure_utc(self._now())
            entries = self._prune(self._load_entries(), now)
            scoped_key = self._scoped_key(source, skill_id, idempotency_key)
            entry = entries.get(scoped_key)
            if entry and entry.expires_at > now:
                self._save_entries(entries)
                return IdempotencyDecision(allowed=False, expires_at=entry.expires_at)

            if ttl_seconds <= 0:
                self._save_entries(entries)
                return IdempotencyDecision(allowed=True, expires_at=None)

            expires_at = now + timedelta(seconds=ttl_seconds)
            entries[scoped_key] = IdempotencyEntry(
                source=source,
                skill_id=skill_id,
                idempotency_key=idempotency_key,
                expires_at=expires_at,
            )
            self._save_entries(entries)
            return IdempotencyDecision(allowed=True, expires_at=expires_at)

    def _scoped_key(self, source: str, skill_id: str, idempotency_key: str) -> str:
        return f"{source}:{skill_id}:{idempotency_key}"

    def _load_entries(self) -> dict[str, IdempotencyEntry]:
        if not self.path.exists():
            return {}
        raw = json.loads(self.path.read_text(encoding="utf-8"))
        entries_raw = raw.get("entries", raw)
        if not isinstance(entries_raw, dict):
            return {}
        entries: dict[str, IdempotencyEntry] = {}
        for scoped_key, payload in entries_raw.items():
            if not isinstance(payload, dict):
                continue
            source = str(payload.get("source", "")).strip()
            skill_id = str(payload.get("skill_id", "")).strip()
            key = str(payload.get("idempotency_key", "")).strip()
            expires_raw = payload.get("expires_at")
            if not source or not skill_id or not key or not expires_raw:
                continue
            entries[scoped_key] = IdempotencyEntry(
                source=source,
                skill_id=skill_id,
                idempotency_key=key,
                expires_at=_parse_datetime(str(expires_raw)),
            )
        return entries

    def _save_entries(self, entries: dict[str, IdempotencyEntry]) -> None:
        payload = {
            "entries": {
                key: entry.to_dict() for key, entry in entries.items()
            }
        }
        atomic_write_text(
            self.path,
            json.dumps(payload, ensure_ascii=True, indent=2),
            encoding="utf-8",
        )

    def _prune(
        self,
        entries: dict[str, IdempotencyEntry],
        now: datetime,
    ) -> dict[str, IdempotencyEntry]:
        current = _ensure_utc(now)
        return {
            key: entry
            for key, entry in entries.items()
            if entry.expires_at > current
        }


class IdempotencyStorePostgres:
    def __init__(
        self,
        dsn: str,
        *,
        tenant_id: str,
        now_provider: Callable[[], datetime] = _utc_now,
        schema: str = "skillos",
    ) -> None:
        self._dsn = dsn
        self._tenant_id = tenant_id
        self._now = now_provider
        self._schema = schema
        self._ensure_table()

    def check_and_record(
        self,
        source: str,
        skill_id: str,
        idempotency_key: str,
        *,
        ttl_seconds: int = DEFAULT_TTL_SECONDS,
    ) -> IdempotencyDecision:
        now = _ensure_utc(self._now())
        expires_at = None
        with pg_connect(self._dsn) as conn:
            conn.execute(
                f"DELETE FROM {self._schema}.idempotency_entries "
                "WHERE tenant_id = %s AND expires_at <= %s",
                (self._tenant_id, now),
            )
            row = conn.execute(
                f"""
                SELECT expires_at FROM {self._schema}.idempotency_entries
                WHERE tenant_id = %s AND source = %s AND skill_id = %s
                AND idempotency_key = %s
                """,
                (self._tenant_id, source, skill_id, idempotency_key),
            ).fetchone()
            if row and row.get("expires_at") and row["expires_at"] > now:
                return IdempotencyDecision(allowed=False, expires_at=row["expires_at"])
            if ttl_seconds <= 0:
                return IdempotencyDecision(allowed=True, expires_at=None)
            expires_at = now + timedelta(seconds=ttl_seconds)
            conn.execute(
                f"""
                INSERT INTO {self._schema}.idempotency_entries
                    (tenant_id, source, skill_id, idempotency_key, expires_at)
                VALUES (%s, %s, %s, %s, %s)
                ON CONFLICT (tenant_id, source, skill_id, idempotency_key)
                DO UPDATE SET expires_at = EXCLUDED.expires_at
                """,
                (self._tenant_id, source, skill_id, idempotency_key, expires_at),
            )
        return IdempotencyDecision(allowed=True, expires_at=expires_at)

    def _ensure_table(self) -> None:
        with pg_connect(self._dsn) as conn:
            conn.execute(f"CREATE SCHEMA IF NOT EXISTS {self._schema}")
            conn.execute(
                f"""
                CREATE TABLE IF NOT EXISTS {self._schema}.idempotency_entries (
                    tenant_id TEXT NOT NULL,
                    source TEXT NOT NULL,
                    skill_id TEXT NOT NULL,
                    idempotency_key TEXT NOT NULL,
                    expires_at TIMESTAMPTZ NOT NULL,
                    PRIMARY KEY (tenant_id, source, skill_id, idempotency_key)
                )
                """
            )
            conn.execute(
                f"""
                CREATE INDEX IF NOT EXISTS idempotency_expires_idx
                ON {self._schema}.idempotency_entries (tenant_id, expires_at)
                """
            )


def idempotency_store_from_env(root: Path) -> IdempotencyStore | IdempotencyStorePostgres:
    config = storage_backend_from_env()
    if config.backend != "postgres":
        return IdempotencyStore(default_idempotency_path(root))
    dsn = require_postgres_dsn(config, context="idempotency")
    tenant_id = resolve_tenant_id(Path(root))
    return IdempotencyStorePostgres(
        dsn,
        tenant_id=tenant_id,
        schema=config.postgres_schema,
    )



from __future__ import annotations

from concurrent.futures import ThreadPoolExecutor, as_completed
from dataclasses import dataclass
import inspect
import json
import os
from pathlib import Path
import time
from typing import Iterable

import yaml

from skillos.approval_gate import ApprovalDecision, ApprovalGate
from skillos.authorization import PermissionChecker, default_permissions_path
from skillos.budget import (
    BudgetManager,
    budget_config_from_env,
    budget_usage_store_from_env,
)
from skillos.circuit_breaker import (
    CircuitBreaker,
    CircuitBreakerStore,
    circuit_breaker_config_from_env,
    default_circuit_breaker_path,
)
from skillos.policy_engine import ApprovalRequirement, PolicyEngine, default_policy_path
from skillos.risk_scorer import RiskScorer
from skillos.skills.models import SkillMetadata
from skillos.skills.registry import resolve_entrypoint
from skillos.storage import atomic_write_text
from skillos.telemetry import EventLogger, default_log_path, new_request_id
from skillos.tenancy import resolve_tenant_root
from skillos.tool_wrapper import ToolWrapper


COMPOSITION_ENTRYPOINT = "skillos.composition:run_composed_skill"
DEFAULT_PARALLEL_LIMIT = 4

StepGroup = list[str]
StepInput = str | list[str] | tuple[str, ...]


class CompositionError(ValueError):
    """Raised when a composition definition is invalid."""


class CompositionBlocked(CompositionError):
    def __init__(self, reason: str) -> None:
        super().__init__(reason)
        self.reason = reason


@dataclass(frozen=True)
class CompositionSpec:
    skill_id: str
    steps: list[StepGroup]
    version: str
    active: bool = False


@dataclass(frozen=True)
class CompositionActivationResult:
    skill_id: str
    activated: bool
    approval: ApprovalDecision
    status: str


class CompositionStore:
    def __init__(self, root: Path) -> None:
        self._root = Path(root)

    def exists(self, skill_id: str) -> bool:
        return self.path_for(skill_id).exists()

    def load(self, skill_id: str) -> CompositionSpec | None:
        path = self.path_for(skill_id)
        if not path.exists():
            return None
        raw = json.loads(path.read_text(encoding="utf-8"))
        if not isinstance(raw, dict):
            raise CompositionError("Composition file must be a mapping")
        steps = raw.get("steps")
        if not isinstance(steps, list):
            raise CompositionError(
                "Composition steps must be a list of skill ids or parallel groups"
            )
        normalized_steps = _normalize_step_groups(steps)
        version = str(raw.get("version", "0.1.0"))
        active = bool(raw.get("active", False))
        return CompositionSpec(
            skill_id=skill_id,
            steps=normalized_steps,
            version=version,
            active=active,
        )

    def save(self, spec: CompositionSpec) -> None:
        path = self.path_for(spec.skill_id)
        payload = {
            "id": spec.skill_id,
            "steps": _serialize_step_groups(spec.steps),
            "version": spec.version,
            "active": spec.active,
        }
        atomic_write_text(
            path,
            json.dumps(payload, ensure_ascii=True, indent=2),
            encoding="utf-8",
        )

    def path_for(self, skill_id: str) -> Path:
        domain, name = _parse_skill_id(skill_id)
        return self._root / "compositions" / domain / f"{name}.json"


class CompositionEngine:
    def __init__(
        self,
        registry,
        store: CompositionStore | None = None,
        *,
        parallel_limit: int | None = None,
    ) -> None:
        self._registry = registry
        self._store = store or CompositionStore(registry.root)
        self._parallel_limit = parallel_limit or parallel_limit_from_env()
        self._budget_manager = BudgetManager(
            budget_usage_store_from_env(self._registry.root),
            budget_config_from_env(),
        )
        self._permission_checker = PermissionChecker.from_path(
            default_permissions_path(self._registry.root)
        )
        policy_engine = PolicyEngine.from_path(default_policy_path(self._registry.root))
        self._tool_wrapper = ToolWrapper(
            risk_scorer=RiskScorer(),
            policy_engine=policy_engine,
            approval_gate=ApprovalGate(
                required_token=os.getenv("SKILLOS_APPROVAL_TOKEN")
            ),
        )
        self._circuit_breaker = CircuitBreaker(
            CircuitBreakerStore(default_circuit_breaker_path(self._registry.root)),
            circuit_breaker_config_from_env(),
        )

    def execute(
        self,
        skill_id: str,
        payload: str,
        *,
        allow_inactive: bool = False,
        _stack: list[str] | None = None,
        logger: EventLogger | None = None,
        role: str | None = None,
        attributes: dict[str, object] | None = None,
        approval_status: str | None = None,
        approval_token: str | None = None,
        charge_budget: bool = True,
        session_context: dict[str, object] | None = None,
    ) -> object:
        stack = list(_stack or [])
        if skill_id in stack:
            raise CompositionError("Composition cycle detected")
        spec = self._store.load(skill_id)
        if not spec:
            metadata = self._registry.get(skill_id)
            if not metadata:
                raise KeyError(f"Unknown skill: {skill_id}")
            from skillos.kernel import get_kernel
            kernel = get_kernel(metadata, root_path=str(self._registry.root))
            return kernel.execute(
                metadata,
                payload,
                role=role,
                attributes=attributes,
                approval_status=approval_status,
                approval_token=approval_token,
                root=self._registry.root,
                charge_budget=charge_budget,
                session_context=session_context,
            )
        if not spec.active and not allow_inactive:
            raise PermissionError(f"Composed skill {skill_id} is not active")
        stack.append(skill_id)
        logger = logger or EventLogger(
            default_log_path(self._registry.root), request_id=new_request_id()
        )
        result: object = payload
        order_index = 0
        for group_index, group in enumerate(spec.steps):
            group_payload = str(result)
            if len(group) == 1:
                result = self._execute_step(
                    skill_id,
                    group[0],
                    group_payload,
                    allow_inactive=allow_inactive,
                    stack=stack,
                    logger=logger,
                    order_index=order_index,
                    group_index=group_index,
                    role=role,
                    attributes=attributes,
                    approval_status=approval_status,
                    approval_token=approval_token,
                    charge_budget=charge_budget,
                    session_context=session_context,
                )
            else:
                result = self._execute_parallel_group(
                    skill_id,
                    group,
                    group_payload,
                    allow_inactive=allow_inactive,
                    stack=stack,
                    logger=logger,
                    order_index=order_index,
                    group_index=group_index,
                    role=role,
                    attributes=attributes,
                    approval_status=approval_status,
                    approval_token=approval_token,
                    charge_budget=charge_budget,
                    session_context=session_context,
                )
            order_index += len(group)
        return result

    def _execute_step(
        self,
        composed_skill_id: str,
        step_id: str,
        payload: str,
        *,
        allow_inactive: bool,
        stack: list[str],
        logger: EventLogger,
        order_index: int,
        group_index: int,
        role: str | None,
        attributes: dict[str, object] | None,
        approval_status: str | None,
        approval_token: str | None,
        charge_budget: bool,
        session_context: dict[str, object] | None,
    ) -> object:
        start = time.perf_counter()
        try:
            self._authorize_step(
                step_id,
                payload,
                role=role,
                attributes=attributes,
                approval_status=approval_status,
                approval_token=approval_token,
                charge_budget=charge_budget,
            )
            output = self.execute(
                step_id,
                payload=payload,
                allow_inactive=allow_inactive,
                _stack=stack,
                logger=logger,
                role=role,
                attributes=attributes,
                approval_status=approval_status,
                approval_token=approval_token,
                charge_budget=charge_budget,
                session_context=session_context,
            )
        except CompositionBlocked as exc:
            duration_ms = (time.perf_counter() - start) * 1000
            _log_composition_step(
                logger,
                composed_skill_id,
                step_id,
                order_index,
                group_index,
                duration_ms,
                status="blocked",
                error_class=exc.reason,
            )
            raise
        except Exception as exc:
            duration_ms = (time.perf_counter() - start) * 1000
            _log_composition_step(
                logger,
                composed_skill_id,
                step_id,
                order_index,
                group_index,
                duration_ms,
                status="error",
                error_class=exc.__class__.__name__,
            )
            raise
        duration_ms = (time.perf_counter() - start) * 1000
        _log_composition_step(
            logger,
            composed_skill_id,
            step_id,
            order_index,
            group_index,
            duration_ms,
            status="success",
        )
        return output

    def _execute_parallel_group(
        self,
        composed_skill_id: str,
        group: list[str],
        payload: str,
        *,
        allow_inactive: bool,
        stack: list[str],
        logger: EventLogger,
        order_index: int,
        group_index: int,
        role: str | None,
        attributes: dict[str, object] | None,
        approval_status: str | None,
        approval_token: str | None,
        charge_budget: bool,
        session_context: dict[str, object] | None,
    ) -> str:
        missing = object()
        outputs: list[object] = [missing for _ in group]
        durations: list[float | None] = [None for _ in group]
        for step_id in group:
            self._authorize_step(
                step_id,
                payload,
                role=role,
                attributes=attributes,
                approval_status=approval_status,
                approval_token=approval_token,
                charge_budget=charge_budget,
            )
        with ThreadPoolExecutor(max_workers=self._parallel_limit) as executor:
            future_map = {}
            for index, step_id in enumerate(group):
                future = executor.submit(
                    self._execute_with_timing,
                    step_id,
                    payload,
                    allow_inactive,
                    stack,
                    logger,
                    durations,
                    index,
                    role,
                    attributes,
                    approval_status,
                    approval_token,
                    charge_budget,
                    session_context,
                )
                future_map[future] = (index, step_id)

            for future in as_completed(future_map):
                index, step_id = future_map[future]
                try:
                    output = future.result()
                except Exception as exc:
                    duration_ms = durations[index] or 0.0
                    _log_composition_step(
                        logger,
                        composed_skill_id,
                        step_id,
                        order_index + index,
                        group_index,
                        duration_ms,
                        status="error",
                        error_class=exc.__class__.__name__,
                    )
                    for pending in future_map:
                        if pending is not future:
                            pending.cancel()
                    raise
                duration_ms = durations[index] or 0.0
                outputs[index] = output
                _log_composition_step(
                    logger,
                    composed_skill_id,
                    step_id,
                    order_index + index,
                    group_index,
                    duration_ms,
                    status="success",
                )
        return "\n".join(str(output) for output in outputs if output is not missing)

    def _execute_with_timing(
        self,
        step_id: str,
        payload: str,
        allow_inactive: bool,
        stack: list[str],
        logger: EventLogger,
        durations: list[float | None],
        index: int,
        role: str | None,
        attributes: dict[str, object] | None,
        approval_status: str | None,
        approval_token: str | None,
        charge_budget: bool,
        session_context: dict[str, object] | None,
    ) -> object:
        start = time.perf_counter()
        try:
            return self.execute(
                step_id,
                payload=payload,
                allow_inactive=allow_inactive,
                _stack=stack,
                logger=logger,
                role=role,
                attributes=attributes,
                approval_status=approval_status,
                approval_token=approval_token,
                charge_budget=charge_budget,
                session_context=session_context,
            )
        finally:
            durations[index] = (time.perf_counter() - start) * 1000

    def _authorize_step(
        self,
        step_id: str,
        payload: str,
        *,
        role: str | None,
        attributes: dict[str, object] | None,
        approval_status: str | None,
        approval_token: str | None,
        charge_budget: bool,
    ) -> None:
        metadata = self._registry.get(step_id)
        skill_tags = metadata.tags if metadata else None
        perm_decision = self._permission_checker.authorize(
            step_id,
            role,
            skill_tags=skill_tags,
            attributes=attributes,
        )
        if not perm_decision.allowed:
            raise CompositionBlocked(perm_decision.policy_id or "permission_denied")

        approval_decision = self._tool_wrapper.authorize(
            step_id,
            payload,
            approval_status=approval_status,
            approval_token=approval_token,
        )
        if not approval_decision.approval.allowed:
            raise CompositionBlocked(
                approval_decision.approval.policy_id or "approval_required"
            )

        circuit_decision = self._circuit_breaker.allow(step_id)
        if not circuit_decision.allowed:
            raise CompositionBlocked(circuit_decision.reason or "circuit_breaker_open")

        budget_result = (
            self._budget_manager.authorize()
            if charge_budget
            else self._budget_manager.evaluate()
        )
        if not budget_result.allowed:
            raise CompositionBlocked(budget_result.reason or "budget_block")


def compose_skill(
    root: Path,
    skill_id: str,
    steps: Iterable[StepInput],
    *,
    name: str | None = None,
    description: str | None = None,
    tags: list[str] | None = None,
) -> CompositionSpec:
    root_path = Path(root)
    normalized_steps = _normalize_step_groups(steps)
    store = CompositionStore(root_path)
    _validate_no_cycles(skill_id, normalized_steps, store)

    from skillos.skills.registry import SkillRegistry

    registry = SkillRegistry(root_path)
    registry.load_all()
    _validate_steps_exist(normalized_steps, registry, store)
    _validate_io_contracts(normalized_steps, registry, store)

    existing = store.load(skill_id)
    metadata_path = _metadata_path(root_path, skill_id)
    if metadata_path.exists() and not existing:
        raise CompositionError("Skill already exists")
    version = _next_version(existing.version) if existing else "0.1.0"
    spec = CompositionSpec(
        skill_id=skill_id,
        steps=normalized_steps,
        version=version,
        active=False,
    )
    store.save(spec)

    _write_metadata(
        metadata_path,
        skill_id,
        name=name,
        description=description,
        tags=tags,
        version=version,
    )
    return spec


def execute_composed_skill(
    skill_id: str,
    registry,
    payload: str,
    *,
    allow_inactive: bool = False,
) -> object:
    engine = CompositionEngine(registry)
    return engine.execute(skill_id, payload, allow_inactive=allow_inactive)


def activate_composed_skill(
    root: Path,
    skill_id: str,
    *,
    approval_status: str | None = None,
    approval_token: str | None = None,
    required_token: str | None = None,
    require_tests: bool = True,
) -> CompositionActivationResult:
    root_path = Path(root)
    store = CompositionStore(root_path)
    spec = store.load(skill_id)
    if not spec:
        raise CompositionError(f"Unknown composed skill: {skill_id}")
    if require_tests and not _coverage_path(root_path, skill_id).exists():
        approval = ApprovalDecision(
            allowed=False,
            policy_id="tests_required",
            status="tests_required",
        )
        return CompositionActivationResult(
            skill_id=skill_id,
            activated=False,
            approval=approval,
            status="tests_required",
        )
    requirement = ApprovalRequirement(
        required=True,
        policy_id="composition_activation",
        reasons=["composition_activation"],
        risk_score=0,
        risk_level="low",
    )
    gate = ApprovalGate(required_token=required_token)
    approval = gate.check(requirement, approval_status, approval_token)
    if not approval.allowed:
        return CompositionActivationResult(
            skill_id=skill_id,
            activated=False,
            approval=approval,
            status=approval.status,
        )
    activated = CompositionSpec(
        skill_id=spec.skill_id,
        steps=spec.steps,
        version=spec.version,
        active=True,
    )
    store.save(activated)
    return CompositionActivationResult(
        skill_id=skill_id,
        activated=True,
        approval=approval,
        status="activated",
    )


def run_composed_skill(payload: str = "ok") -> str:
    raise RuntimeError("Composed skills must be executed via SkillRegistry")


def _metadata_path(root: Path, skill_id: str) -> Path:
    domain, name = _parse_skill_id(skill_id)
    return Path(root) / "metadata" / domain / f"{name}.yaml"


def _write_metadata(
    path: Path,
    skill_id: str,
    *,
    name: str | None,
    description: str | None,
    tags: list[str] | None,
    version: str,
) -> None:
    domain, name_part = _parse_skill_id(skill_id)
    tags = tags or [domain, "composed"]
    metadata = SkillMetadata(
        id=skill_id,
        name=name or name_part.replace("_", " ").title(),
        description=description or f"Composed skill for {skill_id}",
        version=version,
        entrypoint=COMPOSITION_ENTRYPOINT,
        tags=tags,
    )
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(
        yaml.safe_dump(metadata.model_dump(), sort_keys=False),
        encoding="utf-8",
    )


def _coverage_path(root: Path, skill_id: str) -> Path:
    root_path = resolve_tenant_root(root)
    safe_id = skill_id.replace("/", "_")
    return Path(root_path) / "coverage" / f"{safe_id}.json"


def _validate_steps_exist(steps, registry, store: CompositionStore) -> None:
    missing = [
        step
        for step in _iter_step_ids(steps)
        if not store.exists(step) and registry.get(step) is None
    ]
    if missing:
        raise CompositionError(f"Unknown skills: {', '.join(missing)}")


def _validate_io_contracts(steps, registry, store: CompositionStore) -> None:
    for step in _iter_step_ids(steps):
        if store.exists(step):
            continue
        metadata = registry.get(step)
        if not metadata:
            raise CompositionError(f"Unknown skill: {step}")
        func = resolve_entrypoint(metadata.entrypoint, registry.root)
        _validate_signature(step, func)


def _validate_signature(skill_id: str, func) -> None:
    signature = inspect.signature(func)
    params = list(signature.parameters.values())
    if any(param.kind in (param.VAR_POSITIONAL, param.VAR_KEYWORD) for param in params):
        return
    required_positional = [
        param
        for param in params
        if param.kind in (param.POSITIONAL_ONLY, param.POSITIONAL_OR_KEYWORD)
        and param.default is param.empty
    ]
    required_keyword = [
        param
        for param in params
        if param.kind is param.KEYWORD_ONLY and param.default is param.empty
    ]
    if required_keyword:
        if (
            len(required_keyword) == 1
            and required_keyword[0].name == "payload"
            and not required_positional
        ):
            return
        raise CompositionError(f"Invalid IO contract for {skill_id}")
    if len(required_positional) > 1:
        raise CompositionError(f"Invalid IO contract for {skill_id}")


def _validate_no_cycles(
    skill_id: str, steps: list[StepGroup], store: CompositionStore
) -> None:
    flat_steps = list(_iter_step_ids(steps))
    if skill_id in flat_steps:
        raise CompositionError("Composition cycle detected")
    visiting: list[str] = []

    def walk(node: str) -> None:
        if node in visiting:
            raise CompositionError("Composition cycle detected")
        spec = store.load(node)
        if not spec:
            return
        visiting.append(node)
        for step in _iter_step_ids(spec.steps):
            if step == skill_id:
                raise CompositionError("Composition cycle detected")
            walk(step)
        visiting.pop()

    for step in flat_steps:
        walk(step)


def parallel_limit_from_env() -> int:
    raw_value = os.getenv("SKILLOS_PARALLEL_LIMIT", str(DEFAULT_PARALLEL_LIMIT))
    try:
        limit = int(raw_value)
    except ValueError:
        return DEFAULT_PARALLEL_LIMIT
    return max(1, limit)


def _normalize_step_groups(steps: Iterable[StepInput]) -> list[StepGroup]:
    normalized: list[StepGroup] = []
    for step in steps:
        if isinstance(step, str):
            cleaned = step.strip()
            if not cleaned:
                raise CompositionError(
                    "Composition steps must be a list of skill ids or parallel groups"
                )
            normalized.append([cleaned])
            continue
        if not isinstance(step, (list, tuple)):
            raise CompositionError(
                "Composition steps must be a list of skill ids or parallel groups"
            )
        group: list[str] = []
        for entry in step:
            if not isinstance(entry, str):
                raise CompositionError(
                    "Composition steps must be a list of skill ids or parallel groups"
                )
            cleaned = entry.strip()
            if not cleaned:
                raise CompositionError(
                    "Composition steps must be a list of skill ids or parallel groups"
                )
            group.append(cleaned)
        if not group:
            raise CompositionError("Composition requires at least one step")
        normalized.append(group)
    if not normalized:
        raise CompositionError("Composition requires at least one step")
    return normalized


def _serialize_step_groups(steps: list[StepGroup]) -> list[object]:
    if all(len(group) == 1 for group in steps):
        return [group[0] for group in steps]
    payload: list[object] = []
    for group in steps:
        if len(group) == 1:
            payload.append(group[0])
        else:
            payload.append(group)
    return payload


def _iter_step_ids(steps: Iterable[StepGroup]) -> Iterable[str]:
    for group in steps:
        for step in group:
            yield step


def _log_composition_step(
    logger: EventLogger,
    composed_skill_id: str,
    step_id: str,
    order_index: int,
    group_index: int,
    duration_ms: float,
    *,
    status: str,
    error_class: str | None = None,
) -> None:
    payload = {
        "status": status,
        "skill_id": composed_skill_id,
        "step_id": step_id,
        "duration_ms": duration_ms,
        "order": order_index,
        "group": group_index,
    }
    if error_class:
        payload["error_class"] = error_class
    logger.log("composition_step", **payload)


def _next_version(version: str) -> str:
    parts = version.split(".")
    if len(parts) != 3 or not all(part.isdigit() for part in parts):
        return "0.1.0"
    major, minor, patch = (int(part) for part in parts)
    return f"{major}.{minor}.{patch + 1}"


def _parse_skill_id(skill_id: str) -> tuple[str, str]:
    if "/" not in skill_id or skill_id.startswith("/") or skill_id.endswith("/"):
        raise CompositionError("Skill id must be in 'domain/name' format")
    domain, name = skill_id.split("/", 1)
    if not domain or not name:
        raise CompositionError("Skill id must be in 'domain/name' format")
    return domain, name



import asyncio
import contextvars
from datetime import datetime, timezone
from pathlib import Path
import os
import time
from typing import Optional, Dict, Any, List
from unittest.mock import MagicMock

from skillos.authorization import PermissionChecker, default_permissions_path
from skillos.approval_gate import ApprovalGate, approval_token_from_env
from skillos.budget import (
    BudgetCheckResult,
    BudgetManager,
    budget_config_from_env,
    budget_usage_store_from_env,
)
from skillos.debugging import (
    DebugTrace,
    DebugTraceConfig,
    trace_step,
)
from skillos.execution import execute_plan
from skillos.execution_planner import (
    build_execution_plan,
    load_execution_plan,
    save_execution_plan,
)
from skillos.circuit_breaker import (
    CircuitBreaker,
    CircuitBreakerStore,
    circuit_breaker_config_from_env,
    default_circuit_breaker_path,
)
from skillos.feedback import FeedbackTracker, feedback_store_from_env
from skillos.policy_engine import PolicyEngine, default_policy_path
from skillos.risk_scorer import RiskScorer
from skillos.routing import build_router_from_env, to_internal_id
from skillos.routing_cache import routing_cache_from_env
from skillos.mode_selector import split_query
from skillos.pipeline import PipelineRunner
from skillos.skills.deprecation import build_deprecation_warning
from skillos.skills.registry import SkillRegistry
from skillos.skills.paths import default_skills_root
from skillos.telemetry import (
    EventLogger,
    default_log_path,
    hash_query,
    new_request_id,
    route_with_telemetry,
    token_count,
)
from skillos.tool_wrapper import ToolWrapper

_CALL_STACK: contextvars.ContextVar[List[str]] = contextvars.ContextVar("_CALL_STACK", default=[])
MAX_CALL_DEPTH = 10

_SESSION_MAX_TOKENS = 2000
_SESSION_KEEP_LAST = 12
_SESSION_RECENT_MESSAGES = 6
_SESSION_SUMMARY_MAX_CHARS = 1200
_SESSION_SUMMARY_KEY = "history_summary"


def _env_int(name: str, default: int) -> int:
    raw = os.getenv(name)
    if raw is None:
        return default
    try:
        return int(str(raw).strip())
    except ValueError:
        return default


class Orchestrator:
    def __init__(
        self,
        root_path: Path,
        log_path: Optional[Path] = None,
        dev_mode: bool = False,
    ):
        self.root_path = root_path
        self.log_path = log_path or default_log_path(root_path)
        self.dev_mode = dev_mode
        self.registry = SkillRegistry(root_path)
        records = self.registry.load_all()
        
        self.skills_metadata = [record.metadata for record in records.values()]
        
        # In dev mode, we can skip complex components
        if dev_mode:
            self.feedback_tracker = MagicMock()
            self.feedback_tracker.get_confidence.return_value = 1.0 # Return float, not MagicMock
            self.router = build_router_from_env(self.skills_metadata) 
            self.routing_cache = MagicMock()
            self.routing_cache.get.return_value = None
            self.budget_manager = MagicMock()
            allowed_result = BudgetCheckResult(
                allowed=True, 
                remaining_daily=1000.0,
                reason=None,
                model="standard",
                estimated_cost=0.0,
                remaining_monthly=10000.0,
                day_key="2024-01-01",
                month_key="2024-01"
            )
            self.budget_manager.authorize.return_value = allowed_result
            self.budget_manager.evaluate.return_value = allowed_result
            self.policy_engine = MagicMock()
            
            # Use a dummy approval gate that returns serializable objects
            # MagicMock causes JSON serialization errors in logging
            self.approval_gate = MagicMock()
            
            self.tool_wrapper = ToolWrapper(
                risk_scorer=MagicMock(),
                policy_engine=self.policy_engine,
                approval_gate=self.approval_gate,
            )
            # Configure tool_wrapper mock decision to be serializable
            # We override authorize method of the real ToolWrapper instance's dependencies
            # Actually, let's just mock tool_wrapper completely to be safe and simple
            self.tool_wrapper = MagicMock()
            
            # Create a dummy object structure for approval decision that is JSON serializable
            # Using simple Namespace or dicts acting as objects
            from types import SimpleNamespace
            
            approval_decision = SimpleNamespace(
                approval=SimpleNamespace(allowed=True, policy_id="dev_allow", status="approved"),
                requirement=SimpleNamespace(required=False, policy_id="dev_policy"),
                risk=SimpleNamespace(score=0.0, level="low")
            )
            self.tool_wrapper.authorize.return_value = approval_decision
            
            self.circuit_breaker = MagicMock()
            self.circuit_breaker.allow.return_value = MagicMock(allowed=True)
            self._policy_path = Path("dummy")
            self._policy_refresh_token = 0.0
            self._approval_token = None
            
        else:
            self.feedback_tracker = FeedbackTracker(feedback_store_from_env(root_path))
            self.router = build_router_from_env(
                self.skills_metadata,
                confidence_provider=self.feedback_tracker.get_confidence,
            )
            self.routing_cache = routing_cache_from_env(root_path)
            
            self.budget_config = budget_config_from_env()
            self.budget_manager = BudgetManager(
                budget_usage_store_from_env(root_path),
                self.budget_config,
            )
            
            self._policy_path = default_policy_path(root_path)
            self._policy_refresh_token = self._policy_refresh_state()
            self._approval_token = approval_token_from_env()
            self.policy_engine = PolicyEngine.from_path(self._policy_path)
            self.approval_gate = ApprovalGate(required_token=self._approval_token)
            self.tool_wrapper = ToolWrapper(
                risk_scorer=RiskScorer(),
                policy_engine=self.policy_engine,
                approval_gate=self.approval_gate,
            )
            self.circuit_breaker = CircuitBreaker(
                CircuitBreakerStore(default_circuit_breaker_path(root_path)),
                circuit_breaker_config_from_env(),
            )

    @classmethod
    def run_simple(cls, query: str, root_path: Path | str, **kwargs) -> Any:
        """
        Convenience method to run a query in dev mode with minimal setup.
        Useful for testing and local development.
        """
        import os
        from uuid import uuid4
        
        # Ensure SKILLOS_ROOT is set if not already
        original_root = os.environ.get("SKILLOS_ROOT")
        if "SKILLOS_ROOT" not in os.environ:
             os.environ["SKILLOS_ROOT"] = str(root_path)
        
        try:
            orchestrator = cls(Path(root_path), dev_mode=True)
            result = orchestrator.run_query(
                query, 
                request_id=f"dev-{uuid4().hex[:8]}", 
                execute=True, 
                mode="auto",
                **kwargs
            )
            
            if result.get("status") == "success":
                return result.get("output")
            return result
        finally:
            if original_root is None:
                os.environ.pop("SKILLOS_ROOT", None)
            else:
                os.environ["SKILLOS_ROOT"] = original_root

    def run_query(
        self,
        query: str,
        request_id: str = None,
        execute: bool = False,
        dry_run: bool = False,
        approval: Optional[str] = None,
        approval_token: Optional[str] = None,
        role: Optional[str] = None,
        attributes: dict[str, object] | None = None,
        tags: list[str] | None = None,
        mode: str | None = None,
        plan_path: Optional[Path] = None,
        debug_trace: Optional[DebugTrace] = None,
        session_id: Optional[str] = None,
    ) -> Dict[str, Any]:
        self._maybe_reload_registry()
        if not self.dev_mode:
            self._maybe_reload_policies()
            
        stack = _CALL_STACK.get()
        if len(stack) >= MAX_CALL_DEPTH:
            raise RecursionError(f"Max call depth exceeded: {MAX_CALL_DEPTH}")

        request_start, request_id, logger, warnings = self._start_request(
            query,
            request_id=request_id,
            debug_trace=debug_trace,
        )

        # Session Management
        session_store = None
        session = None
        session_context = None
        subject = None
        if attributes:
            subject = attributes.get("subject")
            if subject is not None:
                subject = str(subject).strip() or None
        if session_id:
            from skillos.session.store import SessionStore
            session_store = SessionStore(self.root_path)
            session = session_store.get_session(session_id)
            if not session:
                session = session_store.create_session(session_id=session_id, user_id=subject)
            elif subject:
                if session.user_id and session.user_id != subject:
                    raise PermissionError("session_forbidden")
                if session.user_id is None:
                    session.user_id = subject
            
            session.add_message("user", query)
            self._compact_session(session)
            session_context = self._build_session_context(session)
            session_store.save_session(session)


        normalized_mode = (mode or "single").strip().lower()
        segments = split_query(query, normalized_mode)
        if normalized_mode in {"pipeline", "parallel", "auto"} and len(segments) > 1:
            response = self._run_multi(
                segments,
                normalized_mode,
                payload=query,
                execute=execute,
                dry_run=dry_run,
                approval=approval,
                approval_token=approval_token,
                role=role,
                attributes=attributes,
                tags=tags,
                logger=logger,
                request_id=request_id,
                debug_trace=debug_trace,
                request_start=request_start,
                session_context=session_context,
            )
            return self._finalize_session_response(response, session, session_store)

        # 1. Routing
        result = self._route_query(
            query,
            logger=logger,
            request_id=request_id,
            tags=tags,
            debug_trace=debug_trace,
        )

        if result.status == "no_skill_found":
            response = self._handle_no_skill_found(logger, debug_trace, request_start)
            return self._finalize_session_response(response, session, session_store)

        if result.status == "low_confidence":
            response = self._handle_low_confidence(result, logger, debug_trace, request_start)
            return self._finalize_session_response(response, session, session_store)

        # 2. Plan Creation
        plan, metadata, skill_selected = self._prepare_plan(
            result,
            query,
            warnings,
            logger=logger,
            execute=execute,
            dry_run=dry_run,
        )
        response = self.execute_plan(
            plan,
            execute=execute,
            dry_run=dry_run,
            approval=approval,
            approval_token=approval_token,
            role=role,
            attributes=attributes,
            plan_path=plan_path,
            debug_trace=debug_trace,
            logger=logger,
            request_start=request_start,
            warnings=warnings,
            metadata=metadata,
            skill_selected=skill_selected,
            session_context=session_context,
        )

        return self._finalize_session_response(response, session, session_store)

    async def run_query_async(
        self,
        query: str,
        request_id: str = None,
        execute: bool = False,
        dry_run: bool = False,
        approval: Optional[str] = None,
        approval_token: Optional[str] = None,
        role: Optional[str] = None,
        attributes: dict[str, object] | None = None,
        tags: list[str] | None = None,
        mode: str | None = None,
        plan_path: Optional[Path] = None,
        debug_trace: Optional[DebugTrace] = None,
        session_id: Optional[str] = None,
    ) -> Dict[str, Any]:
        self._maybe_reload_registry()
        if not self.dev_mode:
            self._maybe_reload_policies()
            
        stack = _CALL_STACK.get()
        if len(stack) >= MAX_CALL_DEPTH:
            raise RecursionError(f"Max call depth exceeded: {MAX_CALL_DEPTH}")

        request_start, request_id, logger, warnings = self._start_request(
            query,
            request_id=request_id,
            debug_trace=debug_trace,
        )

        # Session Management
        session_store = None
        session = None
        session_context = None
        subject = None
        if attributes:
            subject = attributes.get("subject")
            if subject is not None:
                subject = str(subject).strip() or None
        if session_id:
            from skillos.session.store import SessionStore
            session_store = SessionStore(self.root_path)
            session = session_store.get_session(session_id)
            if not session:
                session = session_store.create_session(session_id=session_id, user_id=subject)
            elif subject:
                if session.user_id and session.user_id != subject:
                    raise PermissionError("session_forbidden")
                if session.user_id is None:
                    session.user_id = subject
            
            session.add_message("user", query)
            self._compact_session(session)
            session_context = self._build_session_context(session)
            session_store.save_session(session)


        normalized_mode = (mode or "single").strip().lower()
        segments = split_query(query, normalized_mode)
        if normalized_mode in {"pipeline", "parallel", "auto"} and len(segments) > 1:
            response = await self._run_multi_async(
                segments,
                normalized_mode,
                payload=query,
                execute=execute,
                dry_run=dry_run,
                approval=approval,
                approval_token=approval_token,
                role=role,
                attributes=attributes,
                tags=tags,
                logger=logger,
                request_id=request_id,
                debug_trace=debug_trace,
                request_start=request_start,
                session_context=session_context,
            )
            return self._finalize_session_response(response, session, session_store)

        result = await self._route_query_async(
            query,
            logger=logger,
            request_id=request_id,
            tags=tags,
            debug_trace=debug_trace,
        )

        if result.status == "no_skill_found":
            response = self._handle_no_skill_found(logger, debug_trace, request_start)
            return self._finalize_session_response(response, session, session_store)

        if result.status == "low_confidence":
            response = self._handle_low_confidence(result, logger, debug_trace, request_start)
            return self._finalize_session_response(response, session, session_store)

        # 2. Plan Creation
        plan, metadata, skill_selected = self._prepare_plan(
            result,
            query,
            warnings,
            logger=logger,
            execute=execute,
            dry_run=dry_run,
        )
        response = await self.execute_plan_async(
            plan,
            execute=execute,
            dry_run=dry_run,
            approval=approval,
            approval_token=approval_token,
            role=role,
            attributes=attributes,
            plan_path=plan_path,
            debug_trace=debug_trace,
            logger=logger,
            request_start=request_start,
            warnings=warnings,
            metadata=metadata,
            skill_selected=skill_selected,
            session_context=session_context,
        )

        return self._finalize_session_response(response, session, session_store)

    def execute_plan(
        self,
        plan,
        *,
        execute: bool,
        dry_run: bool,
        approval: Optional[str],
        approval_token: Optional[str],
        role: Optional[str],
        attributes: dict[str, object] | None,
        plan_path: Optional[Path],
        debug_trace: Optional[DebugTrace],
        logger: EventLogger,
        request_start: float,
        warnings: list[dict[str, object]] | None = None,
        metadata=None,
        skill_selected: dict[str, object] | None = None,
        session_context: dict[str, object] | None = None,
    ) -> Dict[str, Any]:
        warnings_list = list(warnings) if warnings else []
        if metadata is None:
            metadata = self.registry.get(plan.internal_skill_id)
        if metadata and metadata.deprecated:
            already_warned = any(
                warning.get("code") == "deprecated_skill"
                and warning.get("skill_id") == plan.skill_id
                for warning in warnings_list
            )
            if not already_warned:
                warning = build_deprecation_warning(plan.skill_id, metadata)
                warnings_list.append(warning)
                if execute or dry_run:
                    logger.log(
                        "deprecated_skill_used",
                        skill_id=plan.skill_id,
                        deprecation_reason=metadata.deprecation_reason,
                        replacement_id=metadata.replacement_id,
                    )

        if skill_selected:
            logger.log(
                "skill_selected",
                skill_id=skill_selected.get("skill_id"),
                confidence=skill_selected.get("confidence"),
                status=skill_selected.get("status"),
            )

        if plan_path and plan_path.exists() and not dry_run:
            loaded_plan = load_execution_plan(plan_path)
            if loaded_plan.plan_id != plan.plan_id:
                raise ValueError("plan_mismatch")
            plan = loaded_plan

        # 4. Dry Run / Execution
        if dry_run:
            budget_result = self._budget_check(
                execute=False,
                dry_run=True,
                logger=logger,
                debug_trace=debug_trace,
            )
            if not budget_result.allowed and not self.dev_mode: # Bypass budget block in dev mode (safety net)
                 # Note: in dev_mode budget manager is mocked to always allow, but safety first
                return self._handle_block(
                    "budget_block",
                    budget_result.reason,
                    logger,
                    debug_trace,
                    request_start,
                )
             
            with trace_step(debug_trace, "execution", inputs={"dry_run": True}) as trace_output:
                result_preview = execute_plan(
                    self.registry,
                    plan,
                    dry_run=True,
                    session_context=session_context,
                )
                preview = result_preview.preview
                trace_output.update({"executed": result_preview.executed})

            if plan_path:
                save_execution_plan(plan, plan_path)

            return {
                "status": "dry_run",
                "skill_id": plan.skill_id,
                "preview": preview,
                "plan_path": plan_path,
                "warnings": warnings_list or None,
            }

        if execute:
            # 5. Permission Check
            perm_decision = self._permission_check(
                plan,
                metadata,
                role,
                attributes,
                logger=logger,
                debug_trace=debug_trace,
            )
            
            # ALLOW in dev_mode if policies missing
            if self.dev_mode and not perm_decision.allowed:
                 # In dev mode, we typically want to allow execution unless explicitly tested against
                 # But since we mocked everything, it should pass.
                 pass
            elif not perm_decision.allowed:
                return self._handle_block(
                    perm_decision.policy_id,
                    "permission_denied",
                    logger,
                    debug_trace,
                    request_start,
                )

            # 6. Approval Check
            decision = self._approval_check(
                plan,
                approval,
                approval_token,
                logger=logger,
                debug_trace=debug_trace,
            )
            if not decision.approval.allowed:
                return self._handle_block(
                    decision.approval.policy_id,
                    decision.approval.policy_id,
                    logger,
                    debug_trace,
                    request_start,
                )

            # 6.5 Circuit Check
            circuit_decision = self.circuit_breaker.allow(plan.internal_skill_id)
            if not circuit_decision.allowed:
                logger.log(
                    "policy_decision",
                    allowed=False,
                    policy_id="circuit_breaker_open",
                )
                return self._handle_block(
                    "circuit_breaker_open",
                    circuit_decision.reason or "circuit_breaker_open",
                    logger,
                    debug_trace,
                    request_start,
                )

            # 6.6 Budget Check (charge only after permission/approval/circuit)
            budget_result = self._budget_check(
                execute=True,
                dry_run=False,
                logger=logger,
                debug_trace=debug_trace,
            )
            if not budget_result.allowed:
                return self._handle_block(
                    "budget_block",
                    budget_result.reason,
                    logger,
                    debug_trace,
                    request_start,
                )

            # 7. Execute
            output = self._execute_skill(
                plan,
                logger=logger,
                debug_trace=debug_trace,
                request_start=request_start,
                role=role,
                attributes=attributes,
                approval=approval,
                approval_token=approval_token,
                session_context=session_context,
            )
            return {
                "status": "success",
                "skill_id": plan.skill_id,
                "output": output,
                "plan_id": plan.plan_id if plan_path else None,
                "warnings": warnings_list or None,
            }

        return {
            "status": "routed",
            "skill_id": plan.skill_id,
            "warnings": warnings_list or None,
        }


    def _start_request(
        self,
        query: str,
        *,
        request_id: str | None,
        debug_trace: Optional[DebugTrace],
    ) -> tuple[float, str, EventLogger, list[dict[str, object]]]:
        request_start = time.perf_counter()
        request_id = request_id or new_request_id()
        logger = EventLogger(self.log_path, request_id=request_id)
        warnings: list[dict[str, object]] = []
        log_query = bool(
            debug_trace and debug_trace.config.capture_inputs
        )
        request_payload = {
            "query_hash": hash_query(query),
            "query_length": len(query),
            "token_count": token_count(query),
        }
        if log_query:
            request_payload["query"] = query
        logger.log("request_received", **request_payload)
        return request_start, request_id, logger, warnings

    def _session_limits(self) -> dict[str, int]:
        return {
            "max_tokens": _env_int("SKILLOS_SESSION_MAX_TOKENS", _SESSION_MAX_TOKENS),
            "keep_last": _env_int("SKILLOS_SESSION_KEEP_LAST", _SESSION_KEEP_LAST),
            "recent_messages": _env_int(
                "SKILLOS_SESSION_RECENT_MESSAGES", _SESSION_RECENT_MESSAGES
            ),
            "summary_max_chars": _env_int(
                "SKILLOS_SESSION_SUMMARY_MAX_CHARS", _SESSION_SUMMARY_MAX_CHARS
            ),
        }

    def _summarize_messages(self, messages, max_chars: int) -> str:
        lines: list[str] = []
        for message in messages:
            content = str(message.content).strip().replace("\n", " ")
            if len(content) > 160:
                content = f"{content[:157]}..."
            lines.append(f"{message.role}: {content}")
        summary = " | ".join(line for line in lines if line)
        if len(summary) > max_chars:
            summary = summary[-max_chars:]
        return summary

    def _compact_session(self, session) -> None:
        limits = self._session_limits()
        keep_last = max(1, limits["keep_last"])
        max_tokens = max(1, limits["max_tokens"])
        total_tokens = sum(token_count(msg.content) for msg in session.messages)
        if total_tokens <= max_tokens:
            return
        if len(session.messages) <= keep_last:
            return

        to_summarize = session.messages[:-keep_last]
        summary = self._summarize_messages(
            to_summarize,
            limits["summary_max_chars"],
        )
        if summary:
            existing = session.context.get(_SESSION_SUMMARY_KEY)
            if existing:
                combined = f"{existing}\n{summary}"
            else:
                combined = summary
            if len(combined) > limits["summary_max_chars"]:
                combined = combined[-limits["summary_max_chars"]:]
            session.context[_SESSION_SUMMARY_KEY] = combined
        session.messages = session.messages[-keep_last:]

    def _build_session_context(self, session) -> dict[str, object]:
        limits = self._session_limits()
        recent_count = max(0, limits["recent_messages"])
        recent_messages = session.messages[-recent_count:] if recent_count else []
        return {
            "session_id": session.id,
            "user_id": session.user_id,
            "context": dict(session.context),
            "summary": session.context.get(_SESSION_SUMMARY_KEY),
            "recent_messages": [
                {
                    "role": msg.role,
                    "content": msg.content,
                    "timestamp": msg.timestamp.isoformat(),
                }
                for msg in recent_messages
            ],
        }

    def _maybe_reload_registry(self) -> None:
        records = self.registry.reload_if_changed()
        if records is None:
            return
        self.skills_metadata = [record.metadata for record in records.values()]
        self.router = build_router_from_env(
            self.skills_metadata,
            confidence_provider=self.feedback_tracker.get_confidence,
        )

    def _policy_refresh_state(self) -> float:
        try:
            return self._policy_path.stat().st_mtime
        except OSError:
            return 0.0

    def _maybe_reload_policies(self) -> None:
        token = self._policy_refresh_state()
        approval_token = approval_token_from_env()
        if token == self._policy_refresh_token and approval_token == self._approval_token:
            return
        self._policy_refresh_token = token
        self._approval_token = approval_token
        self.policy_engine = PolicyEngine.from_path(self._policy_path)
        self.approval_gate = ApprovalGate(required_token=self._approval_token)
        self.tool_wrapper = ToolWrapper(
            risk_scorer=RiskScorer(),
            policy_engine=self.policy_engine,
            approval_gate=self.approval_gate,
        )

    def _prepare_plan(
        self,
        result,
        query: str,
        warnings: list[dict[str, object]],
        *,
        logger: EventLogger,
        execute: bool,
        dry_run: bool,
    ):
        internal_id = result.internal_skill_id or to_internal_id(result.skill_id)
        plan = build_execution_plan(result.skill_id, internal_id, query)
        metadata = self._collect_deprecation_warning(
            plan.internal_skill_id,
            result.skill_id,
            warnings,
            logger=logger,
            log=execute or dry_run,
        )
        skill_selected = {
            "skill_id": result.skill_id,
            "confidence": result.confidence,
            "status": result.status,
        }
        return plan, metadata, skill_selected

    def _budget_check(
        self,
        *,
        execute: bool,
        dry_run: bool,
        logger: EventLogger,
        debug_trace: Optional[DebugTrace],
    ) -> BudgetCheckResult:
        with trace_step(debug_trace, "budget_check", inputs={"dry_run": dry_run}) as trace_output:
            budget_result = (
                self.budget_manager.authorize()
                if execute and not dry_run
                else self.budget_manager.evaluate()
            )
            trace_output.update(
                {
                    "allowed": budget_result.allowed,
                    "reason": budget_result.reason,
                    "remaining_daily": budget_result.remaining_daily,
                }
            )

        logger.log(
            "budget_check",
            allowed=budget_result.allowed,
            reason=budget_result.reason,
            model=budget_result.model,
            estimated_cost=budget_result.estimated_cost,
            remaining_daily=budget_result.remaining_daily,
            remaining_monthly=budget_result.remaining_monthly,
        )
        return budget_result

    def _permission_check(
        self,
        plan,
        metadata,
        role: Optional[str],
        attributes: dict[str, object] | None,
        *,
        logger: EventLogger,
        debug_trace: Optional[DebugTrace],
    ):
        permission_checker = PermissionChecker.from_path(
            default_permissions_path(self.root_path)
        )
        skill_tags = metadata.tags if metadata else None
        with trace_step(debug_trace, "permission_check", inputs={"role": role}) as trace_output:
            perm_decision = permission_checker.authorize(
                plan.internal_skill_id,
                role,
                skill_tags=skill_tags,
                attributes=attributes,
            )
            trace_output.update({"allowed": perm_decision.allowed})

        logger.log(
            "permission_decision",
            allowed=perm_decision.allowed,
            policy_id=perm_decision.policy_id,
            role=perm_decision.role,
            skill_id=perm_decision.skill_id,
            required_permissions=perm_decision.required_permissions,
            missing_permissions=perm_decision.missing_permissions,
        )
        return perm_decision

    def _approval_check(
        self,
        plan,
        approval: Optional[str],
        approval_token: Optional[str],
        *,
        logger: EventLogger,
        debug_trace: Optional[DebugTrace],
    ):
        with trace_step(debug_trace, "approval_check", inputs={"approval": approval}) as trace_output:
            decision = self.tool_wrapper.authorize(
                plan.internal_skill_id,
                plan.payload,
                approval_status=approval.lower() if approval else None,
                approval_token=approval_token,
            )
            trace_output.update({"allowed": decision.approval.allowed})

        logger.log(
            "policy_decision",
            allowed=decision.approval.allowed,
            policy_id=decision.approval.policy_id,
            approval_required=decision.requirement.required,
            approval_reason=decision.requirement.policy_id,
            risk_score=decision.risk.score,
            risk_level=decision.risk.level,
            approval_status=decision.approval.status,
        )
        return decision
    def _execute_skill(
        self,
        plan,
        *,
        logger: EventLogger,
        debug_trace: Optional[DebugTrace],
        request_start: float,
        role: Optional[str] = None,
        attributes: dict[str, object] | None = None,
        approval: Optional[str] = None,
        approval_token: Optional[str] = None,
        session_context: dict[str, object] | None = None,
    ) -> str | None:
        internal_id = plan.internal_skill_id
        stack = _CALL_STACK.get()
        if internal_id in stack:
            raise RecursionError(f"Circular dependency detected: {internal_id} in {stack}")
        
        token = _CALL_STACK.set(stack + [internal_id])
        try:
            with trace_step(
                debug_trace,
                "execution",
                inputs={"skill_id": plan.internal_skill_id},
            ) as trace_output:
                execution = execute_plan(
                    self.registry,
                    plan,
                    dry_run=False,
                    role=role,
                    attributes=attributes,
                    approval_status=approval,
                    approval_token=approval_token,
                    session_context=session_context,
                    charge_budget=False,
                )
                trace_output.update(
                    {"status": "success", "output": str(execution.output)}
                )

            duration = (time.perf_counter() - request_start) * 1000
            logger.log(
                "execution_result",
                status="success",
                skill_id=plan.skill_id,
                duration_ms=duration,
            )
            self.circuit_breaker.record_success(plan.internal_skill_id)
            return execution.output
        except Exception as exc:
            duration = (time.perf_counter() - request_start) * 1000
            logger.log(
                "execution_result",
                status="error",
                error_class=exc.__class__.__name__,
                duration_ms=duration,
            )
            self.circuit_breaker.record_failure(plan.internal_skill_id)
            raise exc
        finally:
            _CALL_STACK.reset(token)

    def _route_query(
        self,
        query: str,
        *,
        logger: EventLogger,
        request_id: str,
        tags: list[str] | None,
        debug_trace: Optional[DebugTrace],
    ):
        with trace_step(debug_trace, "route", inputs={"query": query}) as trace_output:
            telemetry = route_with_telemetry(
                query,
                self.router,
                logger,
                request_id,
                tags=tags,
                routing_cache=self.routing_cache,
            )
            result = telemetry.result
            trace_output.update({
                "skill_id": result.skill_id,
                "status": result.status,
                "confidence": result.confidence,
                "alternatives": result.alternatives,
                "routing_latency_ms": telemetry.routing_latency_ms,
            })
        return result

    async def _route_query_async(
        self,
        query: str,
        *,
        logger: EventLogger,
        request_id: str,
        tags: list[str] | None,
        debug_trace: Optional[DebugTrace],
    ):
        with trace_step(debug_trace, "route", inputs={"query": query}) as trace_output:
            telemetry = await asyncio.to_thread(
                route_with_telemetry,
                query,
                self.router,
                logger,
                request_id,
                tags=tags,
                routing_cache=self.routing_cache,
            )
            result = telemetry.result
            trace_output.update({
                "skill_id": result.skill_id,
                "status": result.status,
                "confidence": result.confidence,
                "alternatives": result.alternatives,
                "routing_latency_ms": telemetry.routing_latency_ms,
            })
        return result

    async def execute_plan_async(
        self,
        plan,
        *,
        execute: bool,
        dry_run: bool,
        approval: Optional[str],
        approval_token: Optional[str],
        role: Optional[str],
        attributes: dict[str, object] | None,
        plan_path: Optional[Path],
        debug_trace: Optional[DebugTrace],
        logger: EventLogger,
        request_start: float,
        warnings: list[dict[str, object]] | None = None,
        metadata=None,
        skill_selected: dict[str, object] | None = None,
        session_context: dict[str, object] | None = None,
    ) -> Dict[str, Any]:
        return await asyncio.to_thread(
            self.execute_plan,
            plan,
            execute=execute,
            dry_run=dry_run,
            approval=approval,
            approval_token=approval_token,
            role=role,
            attributes=attributes,
            plan_path=plan_path,
            debug_trace=debug_trace,
            logger=logger,
            request_start=request_start,
            warnings=warnings,
            metadata=metadata,
            skill_selected=skill_selected,
            session_context=session_context,
        )
    def _collect_deprecation_warning(
        self,
        internal_skill_id: str,
        public_skill_id: str,
        warnings: list[dict[str, object]],
        *,
        logger: EventLogger | None = None,
        log: bool = False,
    ):
        metadata = self.registry.get(internal_skill_id)
        if metadata and metadata.deprecated:
            warning = build_deprecation_warning(public_skill_id, metadata)
            warnings.append(warning)
            if logger and log:
                logger.log(
                    "deprecated_skill_used",
                    skill_id=public_skill_id,
                    deprecation_reason=metadata.deprecation_reason,
                    replacement_id=metadata.replacement_id,
                )
        return metadata

    def _run_multi(
        self,
        segments: list[str],
        mode: str,
        *,
        payload: str,
        execute: bool,
        dry_run: bool,
        approval: Optional[str],
        approval_token: Optional[str],
        role: Optional[str],
        attributes: dict[str, object] | None,
        tags: list[str] | None,
        logger: EventLogger,
        request_id: str,
        debug_trace: Optional[DebugTrace],
        request_start: float,
        session_context: dict[str, object] | None,
    ) -> Dict[str, Any]:
        if dry_run:
            return {
                "status": "blocked",
                "reason": "dry_run_not_supported",
                "policy_id": "dry_run_not_supported",
            }

        step_ids: list[str] = []
        warnings: list[dict[str, object]] = []
        for segment in segments:
            result = self._route_query(
                segment,
                logger=logger,
                request_id=request_id,
                tags=tags,
                debug_trace=debug_trace,
            )

            if result.status == "no_skill_found":
                return self._handle_no_skill_found(
                    logger, debug_trace, request_start
                )
            if result.status == "low_confidence":
                return self._handle_low_confidence(
                    result, logger, debug_trace, request_start
                )
            if result.skill_id:
                step_ids.append(result.skill_id)
                internal_id = result.internal_skill_id or to_internal_id(result.skill_id)
                metadata = self.registry.get(internal_id)
                if metadata and metadata.deprecated:
                    warning = build_deprecation_warning(result.skill_id, metadata)
                    warnings.append(warning)
                    if execute:
                        logger.log(
                            "deprecated_skill_used",
                            skill_id=result.skill_id,
                            deprecation_reason=metadata.deprecation_reason,
                            replacement_id=metadata.replacement_id,
                        )

        if not execute:
            return {
                "status": "routed",
                "skill_id": step_ids[0] if step_ids else None,
                "steps": step_ids,
                "warnings": warnings or None,
            }

        runner = PipelineRunner(
            self.root_path,
            log_path=self.log_path,
            step_controller=debug_trace.step_controller if debug_trace else None,
        )
        if mode == "parallel":
            steps: list[object] = [step_ids]
        else:
            steps = step_ids
        pipeline_result = runner.run(
            steps,
            payload=payload,
            approval_status=approval,
            approval_token=approval_token,
            role=role,
            attributes=attributes,
            request_id=request_id,
            session_context=session_context,
        )
        if pipeline_result.status == "blocked":
            return {
                "status": "blocked",
                "reason": pipeline_result.reason or "blocked",
                "policy_id": pipeline_result.reason or "blocked",
                "warnings": pipeline_result.warnings or None,
            }
        if pipeline_result.status != "success":
            return {
                "status": "error",
                "reason": pipeline_result.reason or "pipeline_failed",
                "warnings": pipeline_result.warnings or None,
            }
        return {
            "status": "success",
            "output": pipeline_result.output,
            "steps": step_ids,
            "warnings": pipeline_result.warnings or None,
        }

    async def _run_multi_async(
        self,
        segments: list[str],
        mode: str,
        *,
        payload: str,
        execute: bool,
        dry_run: bool,
        approval: Optional[str],
        approval_token: Optional[str],
        role: Optional[str],
        attributes: dict[str, object] | None,
        tags: list[str] | None,
        logger: EventLogger,
        request_id: str,
        debug_trace: Optional[DebugTrace],
        request_start: float,
        session_context: dict[str, object] | None,
    ) -> Dict[str, Any]:
        if dry_run:
            return {
                "status": "blocked",
                "reason": "dry_run_not_supported",
                "policy_id": "dry_run_not_supported",
            }

        step_ids: list[str] = []
        warnings: list[dict[str, object]] = []
        for segment in segments:
            result = await self._route_query_async(
                segment,
                logger=logger,
                request_id=request_id,
                tags=tags,
                debug_trace=debug_trace,
            )

            if result.status == "no_skill_found":
                return self._handle_no_skill_found(
                    logger, debug_trace, request_start
                )
            if result.status == "low_confidence":
                return self._handle_low_confidence(
                    result, logger, debug_trace, request_start
                )
            if result.skill_id:
                step_ids.append(result.skill_id)
                internal_id = result.internal_skill_id or to_internal_id(result.skill_id)
                self._collect_deprecation_warning(
                    internal_id,
                    result.skill_id,
                    warnings,
                    logger=logger,
                    log=execute,
                )

        if not execute:
            return {
                "status": "routed",
                "skill_id": step_ids[0] if step_ids else None,
                "steps": step_ids,
                "warnings": warnings or None,
            }

        runner = PipelineRunner(
            self.root_path,
            log_path=self.log_path,
            step_controller=debug_trace.step_controller if debug_trace else None,
        )
        if mode == "parallel":
            steps: list[object] = [step_ids]
        else:
            steps = step_ids
        pipeline_result = await runner.run_async(
            steps,
            payload=payload,
            approval_status=approval,
            approval_token=approval_token,
            role=role,
            attributes=attributes,
            request_id=request_id,
            session_context=session_context,
        )
        if pipeline_result.status == "blocked":
            return {
                "status": "blocked",
                "reason": pipeline_result.reason or "blocked",
                "policy_id": pipeline_result.reason or "blocked",
                "warnings": pipeline_result.warnings or None,
            }
        if pipeline_result.status != "success":
            return {
                "status": "error",
                "reason": pipeline_result.reason or "pipeline_failed",
                "warnings": pipeline_result.warnings or None,
            }
        return {
            "status": "success",
            "output": pipeline_result.output,
            "steps": step_ids,
            "warnings": pipeline_result.warnings or None,
        }

    def _handle_no_skill_found(self, logger, trace, start_time):
        duration = (time.perf_counter() - start_time) * 1000
        logger.log("execution_result", status="skipped", error_class="no_skill_found", duration_ms=duration)
        return {"status": "no_skill_found"}

    def _handle_low_confidence(self, result, logger, trace, start_time):
        duration = (time.perf_counter() - start_time) * 1000
        logger.log("execution_result", status="skipped", error_class="low_confidence", duration_ms=duration)
        return {
            "status": "low_confidence",
            "skill_id": result.skill_id,
            "alternatives": result.alternatives
        }

    def _handle_block(self, policy_id, reason, logger, trace, start_time):
        duration = (time.perf_counter() - start_time) * 1000
        logger.log("execution_result", status="blocked", error_class=reason, duration_ms=duration)
        return {"status": "blocked", "reason": reason, "policy_id": policy_id}

    def _finalize_session_response(
        self,
        response: Dict[str, Any],
        session,
        session_store,
    ) -> Dict[str, Any]:
        if session and session_store:
            if response.get("status") == "success":
                session.add_message("assistant", response.get("output", ""))
            self._compact_session(session)
            session_store.save_session(session)
            response["session_id"] = session.id
        return response



from __future__ import annotations
import click
import importlib

class LazyGroup(click.Group):
    def __init__(self, *args, lazy_subcommands: dict[str, tuple[str, str]] | None = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.lazy_subcommands = lazy_subcommands or {}

    def list_commands(self, ctx: click.Context) -> list[str]:
        base = super().list_commands(ctx)
        lazy = sorted(self.lazy_subcommands.keys())
        return sorted(set(base + lazy))

    def get_command(self, ctx: click.Context, cmd_name: str) -> click.Command | None:
        if cmd_name in self.lazy_subcommands:
            return self._lazy_load(cmd_name)
        return super().get_command(ctx, cmd_name)

    def _lazy_load(self, cmd_name: str) -> click.Command:
        import_path, obj_name = self.lazy_subcommands[cmd_name]
        mod = importlib.import_module(import_path)
        return getattr(mod, obj_name)


COMMAND_MAPPING = {
    # Core
    "run": ("skillos.cli.commands.core", "run_query"),
    
    # Skills
    "add-skill": ("skillos.cli.commands.skills", "add_skill"),
    "run-skill": ("skillos.cli.commands.skills", "run_skill"),
    "validate": ("skillos.cli.commands.skills", "validate_command"),
    "test": ("skillos.cli.commands.skills", "test_skill"),
    "eval-skill": ("skillos.cli.commands.skills", "eval_skill"),
    "deprecate-skill": ("skillos.cli.commands.skills", "deprecate_skill_command"),
    "undeprecate-skill": ("skillos.cli.commands.skills", "undeprecate_skill_command"),

    # Pipeline
    "pipeline": ("skillos.cli.commands.pipeline", "pipeline"),
    "compose-skill": ("skillos.cli.commands.pipeline", "compose_skill"),
    "activate-skill": ("skillos.cli.commands.pipeline", "activate_skill"),
    
    # Marketplace
    "marketplace": ("skillos.cli.commands.market", "marketplace"),
    
    # Analysis
    "feedback": ("skillos.cli.commands.analysis", "feedback"),
    "optimize": ("skillos.cli.commands.analysis", "optimize_skill"),
    "metrics": ("skillos.cli.commands.analysis", "metrics_report"),
    
    # Infra & Connectors
    "secrets": ("skillos.cli.commands.infra", "secrets"),
    "suggestions": ("skillos.cli.commands.infra", "suggestions"),
    "schedule": ("skillos.cli.commands.infra", "schedule"),
    "webhook": ("skillos.cli.commands.infra", "webhook"),
    "job": ("skillos.cli.commands.infra", "job"),
    "add-connector": ("skillos.cli.commands.infra", "add_connector"),
}


@click.group(cls=LazyGroup, lazy_subcommands=COMMAND_MAPPING)
def cli() -> None:
    """SkillOS command line interface."""


# Preload the run command for test harnesses that access cli.commands directly.
try:
    cli.commands["run"] = cli._lazy_load("run")
except Exception:
    pass


def main() -> None:
    cli()


if __name__ == "__main__":
    main()



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timezone
import json
from pathlib import Path

from skillos.experimentation import ExperimentState, ExperimentStore, VariantMetrics
from skillos.tenancy import resolve_tenant_root


def default_optimization_log_path(root: Path) -> Path:
    root_path = resolve_tenant_root(root)
    return root_path / "experiments" / "optimization.log"


@dataclass(frozen=True)
class OptimizerConfig:
    min_samples: int = 5
    win_rate_delta: float = 0.1
    rollback_delta: float = 0.1


@dataclass(frozen=True)
class OptimizationDecision:
    action: str
    reason: str
    from_variant: str | None
    to_variant: str | None
    metrics: dict[str, object]


class OptimizationLogger:
    def __init__(self, path: Path) -> None:
        self.path = Path(path)

    def log(self, experiment_id: str, decision: OptimizationDecision) -> None:
        self.path.parent.mkdir(parents=True, exist_ok=True)
        payload = {
            "timestamp": datetime.now(timezone.utc).isoformat(),
            "experiment_id": experiment_id,
            "action": decision.action,
            "reason": decision.reason,
            "from_variant": decision.from_variant,
            "to_variant": decision.to_variant,
            "metrics": decision.metrics,
        }
        with self.path.open("a", encoding="utf-8") as handle:
            handle.write(json.dumps(payload, ensure_ascii=True) + "\n")


class Optimizer:
    def __init__(
        self,
        store: ExperimentStore,
        logger: OptimizationLogger,
        config: OptimizerConfig | None = None,
    ) -> None:
        self._store = store
        self._logger = logger
        self._config = config or OptimizerConfig()

    def evaluate(self, experiment_id: str) -> OptimizationDecision:
        state = self._store.get(experiment_id)
        if not state:
            raise ValueError(f"Unknown experiment: {experiment_id}")

        regression_decision = self._check_regression(state)
        if regression_decision:
            self._apply_decision(experiment_id, state, regression_decision)
            return regression_decision

        promotion_decision = self._check_promotion(state)
        if promotion_decision:
            self._apply_decision(experiment_id, state, promotion_decision)
            return promotion_decision

        return OptimizationDecision(
            action="no_action",
            reason="insufficient_evidence",
            from_variant=None,
            to_variant=None,
            metrics=self._snapshot_metrics(state),
        )

    def _check_promotion(
        self, state: ExperimentState
    ) -> OptimizationDecision | None:
        active_metrics = state.variants.get(state.active_variant)
        if not active_metrics or active_metrics.total < self._config.min_samples:
            return None

        best_variant = None
        best_metrics = None
        best_delta = 0.0
        for variant_id, metrics in state.variants.items():
            if variant_id == state.active_variant:
                continue
            if metrics.total < self._config.min_samples:
                continue
            delta = metrics.success_rate - active_metrics.success_rate
            if delta >= self._config.win_rate_delta and delta > best_delta:
                best_variant = variant_id
                best_metrics = metrics
                best_delta = delta

        if not best_variant or not best_metrics:
            return None

        return OptimizationDecision(
            action="promote",
            reason="confidence_threshold_met",
            from_variant=state.active_variant,
            to_variant=best_variant,
            metrics={
                "active": _variant_payload(state.active_variant, active_metrics),
                "candidate": _variant_payload(best_variant, best_metrics),
                "delta": round(best_delta, 4),
            },
        )

    def _check_regression(
        self, state: ExperimentState
    ) -> OptimizationDecision | None:
        if not state.previous_variant:
            return None
        active_metrics = state.variants.get(state.active_variant)
        previous_metrics = state.variants.get(state.previous_variant)
        if not active_metrics or not previous_metrics:
            return None
        if (
            active_metrics.total < self._config.min_samples
            or previous_metrics.total < self._config.min_samples
        ):
            return None
        delta = previous_metrics.success_rate - active_metrics.success_rate
        if delta < self._config.rollback_delta:
            return None
        return OptimizationDecision(
            action="rollback",
            reason="regression_detected",
            from_variant=state.active_variant,
            to_variant=state.previous_variant,
            metrics={
                "active": _variant_payload(state.active_variant, active_metrics),
                "previous": _variant_payload(state.previous_variant, previous_metrics),
                "delta": round(delta, 4),
            },
        )

    def _apply_decision(
        self,
        experiment_id: str,
        state: ExperimentState,
        decision: OptimizationDecision,
    ) -> None:
        if decision.action == "promote" and decision.to_variant:
            state.previous_variant = state.active_variant
            state.active_variant = decision.to_variant
        elif decision.action == "rollback" and decision.to_variant:
            state.active_variant = decision.to_variant
            state.previous_variant = None
        self._store.upsert(experiment_id, state)
        self._logger.log(experiment_id, decision)

    def _snapshot_metrics(self, state: ExperimentState) -> dict[str, object]:
        return {
            variant_id: _variant_payload(variant_id, metrics)
            for variant_id, metrics in state.variants.items()
        }


def _variant_payload(variant_id: str, metrics: VariantMetrics) -> dict[str, object]:
    return {
        "variant": variant_id,
        "successes": metrics.successes,
        "failures": metrics.failures,
        "total": metrics.total,
        "success_rate": round(metrics.success_rate, 4),
    }



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

from skillos.skills.errors import SkillValidationError
from skillos.skills.loader import load_skill_file
from skillos.skills.registry import resolve_entrypoint


@dataclass(frozen=True)
class SkillValidationIssue:
    path: Path
    category: str
    message: str
    skill_id: str | None = None


def validate_skills(
    root: Path, *, check_entrypoints: bool = True
) -> list[SkillValidationIssue]:
    root_path = Path(root)
    metadata_path = root_path / "metadata"
    if not metadata_path.exists():
        return []

    issues: list[SkillValidationIssue] = []
    for file_path in _iter_skill_files(metadata_path):
        try:
            metadata = load_skill_file(file_path)
        except SkillValidationError as exc:
            issues.append(
                SkillValidationIssue(
                    path=file_path,
                    category="metadata",
                    message=str(exc),
                )
            )
            continue

        if not check_entrypoints:
            continue

        try:
            resolve_entrypoint(metadata.entrypoint, root_path)
        except Exception as exc:  # pragma: no cover - defensive
            issues.append(
                SkillValidationIssue(
                    path=file_path,
                    category="entrypoint",
                    message=_format_entrypoint_error(exc),
                    skill_id=metadata.id,
                )
            )
    return issues


def _iter_skill_files(metadata_path: Path) -> list[Path]:
    yaml_files = list(metadata_path.rglob("*.yaml"))
    yml_files = list(metadata_path.rglob("*.yml"))
    return sorted(yaml_files + yml_files)


def _format_entrypoint_error(exc: Exception) -> str:
    message = str(exc).strip()
    if message:
        return f"{exc.__class__.__name__}: {message}"
    return exc.__class__.__name__



"""Skill management helpers."""

from skillos.skills.errors import SkillValidationError
from skillos.skills.loader import load_skill_file
from skillos.skills.eval import (
    EvalCaseResult,
    EvalRunResult,
    SkillEvalError,
    default_eval_result_path,
    run_skill_eval,
    save_eval_result,
)
from skillos.skills.models import SkillEvalCase, SkillEvalConfig, SkillMetadata
from skillos.skills.registry import SkillRegistry

__all__ = [
    "EvalCaseResult",
    "EvalRunResult",
    "SkillEvalCase",
    "SkillEvalConfig",
    "SkillEvalError",
    "SkillMetadata",
    "SkillRegistry",
    "SkillValidationError",
    "default_eval_result_path",
    "load_skill_file",
    "run_skill_eval",
    "save_eval_result",
]



from __future__ import annotations

from contextlib import contextmanager
import os
import tempfile
from pathlib import Path
import time
from typing import Iterator

try:
    import msvcrt

    def _lock(fd: int) -> None:
        # LK_NBLCK is non-blocking on Windows
        msvcrt.locking(fd, msvcrt.LK_NBLCK, 1)

    def _unlock(fd: int) -> None:
        msvcrt.locking(fd, msvcrt.LK_UNLCK, 1)

except ImportError:
    # Fallback for non-Windows (simplified, typically use fcntl)
    try:
        import fcntl

        def _lock(fd: int) -> None:
            fcntl.flock(fd, fcntl.LOCK_EX)

        def _unlock(fd: int) -> None:
            fcntl.flock(fd, fcntl.LOCK_UN)

    except ImportError:  # pragma: no cover
        def _lock(fd: int) -> None:
            pass

        def _unlock(fd: int) -> None:
            pass


@contextmanager
def file_lock(path: Path, timeout: float = 5.0) -> Iterator[None]:
    """Simple cross-platform file lock."""
    lock_path = path.with_suffix(path.suffix + ".lock")
    start_time = time.time()
    
    # Ensure directory exists
    lock_path.parent.mkdir(parents=True, exist_ok=True)
    
    while True:
        try:
            # Open for writing (creates if not exists)
            fd = os.open(lock_path, os.O_RDWR | os.O_CREAT)
            try:
                _lock(fd)
                yield
                try:
                    _unlock(fd)
                except (IOError, OSError):
                    pass # Best effort unlock before close
                break
            finally:
                os.close(fd)
        except (IOError, OSError):
            if time.time() - start_time > timeout:
                raise TimeoutError(f"Could not acquire lock on {path} after {timeout}s")
            time.sleep(0.1)


def atomic_write_text(path: Path, data: str, *, encoding: str = "utf-8") -> None:
    path = Path(path)
    path.parent.mkdir(parents=True, exist_ok=True)
    fd, temp_name = tempfile.mkstemp(
        dir=path.parent,
        prefix=f".{path.name}.",
        suffix=".tmp",
    )
    try:
        with os.fdopen(fd, "w", encoding=encoding) as handle:
            handle.write(data)
        os.replace(temp_name, path)
    finally:
        try:
            os.unlink(temp_name)
        except FileNotFoundError:
            pass



from __future__ import annotations

from dataclasses import dataclass
import hashlib
import hmac
import json
import os
from pathlib import Path
import time

from skillos.attachments import AttachmentError, AttachmentReference, ingest_attachments
from skillos.connectors import SecretResolutionError, SecretResolver, SecretsStore, default_secrets_path
from skillos.idempotency import (
    IdempotencyDecision,
    idempotency_store_from_env,
    idempotency_ttl_from_env,
)
from skillos.jobs import job_store_from_env
from skillos.routing import to_internal_id
from skillos.telemetry import (
    EventLogger,
    default_log_path,
    log_job_enqueued,
    new_request_id,
)
from skillos.tenancy import resolve_tenant_root


class WebhookError(ValueError):
    pass


class WebhookTriggerError(WebhookError):
    pass


class WebhookPayloadError(WebhookError):
    pass


class WebhookSignatureError(WebhookError):
    def __init__(self, message: str, status_code: int) -> None:
        super().__init__(message)
        self.status_code = status_code


@dataclass(frozen=True)
class WebhookTrigger:
    trigger_id: str
    skill_id: str


@dataclass(frozen=True)
class WebhookHandleResult:
    status: str
    trigger_id: str
    skill_id: str
    job_id: str | None
    idempotency_key: str | None
    expires_at: str | None


DEFAULT_SIGNATURE_TTL_SECONDS = 300


def default_webhook_path(root: Path) -> Path:
    root_path = resolve_tenant_root(root)
    return root_path / "triggers" / "webhooks.json"


def load_webhook_triggers(path: Path) -> dict[str, WebhookTrigger]:
    if not path.exists():
        return {}
    raw = json.loads(path.read_text(encoding="utf-8"))
    items = raw.get("webhooks", raw.get("triggers", raw))
    if not isinstance(items, list):
        raise WebhookTriggerError("invalid_trigger_format")
    triggers: dict[str, WebhookTrigger] = {}
    for item in items:
        if not isinstance(item, dict):
            continue
        trigger_id = str(item.get("id") or item.get("trigger_id") or "").strip()
        skill_id = str(item.get("skill_id") or "").strip()
        if not trigger_id or not skill_id:
            continue
        triggers[trigger_id] = WebhookTrigger(
            trigger_id=trigger_id,
            skill_id=skill_id,
        )
    return triggers


def resolve_webhook_trigger(path: Path, trigger_id: str) -> WebhookTrigger:
    triggers = load_webhook_triggers(path)
    trigger = triggers.get(trigger_id)
    if trigger is None:
        raise WebhookTriggerError("trigger_not_found")
    return trigger


def load_webhook_payload(path: Path) -> tuple[dict[str, object], str]:
    raw = path.read_text(encoding="utf-8")
    payload = _parse_webhook_payload(raw)
    return payload, raw


def _parse_webhook_payload(raw: str) -> dict[str, object]:
    try:
        payload = json.loads(raw)
    except json.JSONDecodeError as exc:
        raise WebhookPayloadError("invalid_payload") from exc
    if not isinstance(payload, dict):
        raise WebhookPayloadError("invalid_payload")
    return payload


def handle_webhook_event(
    trigger_id: str,
    payload_path: Path,
    root_path: Path,
    *,
    signature: str | None = None,
    idempotency_key: str | None = None,
    log_path: Path | None = None,
    ttl_seconds: int | None = None,
    signature_ttl_seconds: int | None = None,
) -> WebhookHandleResult:
    trigger = resolve_webhook_trigger(default_webhook_path(root_path), trigger_id)
    raw_payload = payload_path.read_text(encoding="utf-8")
    request_id = new_request_id()
    logger = EventLogger(log_path or default_log_path(root_path), request_id=request_id)

    try:
        _validate_signature(
            signature,
            root_path,
            raw_payload,
            ttl_seconds=signature_ttl_seconds,
        )
    except WebhookSignatureError as exc:
        logger.log(
            "webhook_received",
            trigger_id=trigger.trigger_id,
            skill_id=trigger.skill_id,
            status="rejected",
            status_code=exc.status_code,
            reason=str(exc),
        )
        raise

    from skillos.rate_limit import rate_limiter_from_env
    from skillos.storage_backend import resolve_tenant_id
    
    tenant_id = resolve_tenant_id(root_path)
    limiter = rate_limiter_from_env()
    # Scoped to tenant to prevent collision
    rl_key = f"tenant:{tenant_id}:webhook:{trigger_id}"
    
    if not limiter.check_and_consume(rl_key):
        logger.log(
            "webhook_received",
            trigger_id=trigger.trigger_id,
            skill_id=trigger.skill_id,
            status="rate_limited",
            status_code=429,
        )
        raise WebhookTriggerError("rate_limit_exceeded")

    try:
        payload = _parse_webhook_payload(raw_payload)
    except WebhookPayloadError as exc:
        logger.log(
            "webhook_received",
            trigger_id=trigger.trigger_id,
            skill_id=trigger.skill_id,
            status="rejected",
            status_code=400,
            reason=str(exc),
        )
        raise

    logger.log(
        "webhook_received",
        trigger_id=trigger.trigger_id,
        skill_id=trigger.skill_id,
        status="accepted",
        status_code=200,
    )

    resolved_key = _resolve_idempotency_key(payload, idempotency_key)
    decision = None
    if resolved_key:
        decision = _check_idempotency(
            resolved_key,
            trigger,
            root_path,
            ttl_seconds=ttl_seconds,
        )
        if not decision.allowed:
            expires_at = decision.expires_at.isoformat() if decision.expires_at else None
            logger.log(
                "idempotency_skipped",
                source="webhook",
                skill_id=trigger.skill_id,
                idempotency_key=resolved_key,
                expires_at=expires_at,
            )
            return WebhookHandleResult(
                status="skipped",
                trigger_id=trigger.trigger_id,
                skill_id=trigger.skill_id,
                job_id=None,
                idempotency_key=resolved_key,
                expires_at=expires_at,
            )

    try:
        attachments = ingest_attachments(
            payload.get("attachments"),
            root_path,
            request_id=request_id,
        )
    except AttachmentError as exc:
        raise WebhookPayloadError(str(exc)) from exc

    if attachments:
        logger.log(
            "attachments_ingested",
            count=len(attachments),
            total_bytes=sum(attachment.size_bytes for attachment in attachments),
            attachments=[attachment.to_dict() for attachment in attachments],
        )

    job_payload = _resolve_job_payload(payload, raw_payload, attachments)
    store = job_store_from_env(root_path)
    record = store.enqueue(
        to_internal_id(trigger.skill_id),
        payload=job_payload,
        max_retries=0,
    )
    log_job_enqueued(
        logger,
        job_id=record.job_id,
        skill_id=record.skill_id,
        status=record.status,
        max_retries=record.max_retries,
    )
    expires_at = decision.expires_at.isoformat() if decision and decision.expires_at else None
    return WebhookHandleResult(
        status="enqueued",
        trigger_id=trigger.trigger_id,
        skill_id=trigger.skill_id,
        job_id=record.job_id,
        idempotency_key=resolved_key,
        expires_at=expires_at,
    )


def build_signature_header(
    secret: str, raw_body: str, *, timestamp: int | None = None
) -> str:
    ts = int(timestamp if timestamp is not None else time.time())
    signature = compute_signature(secret, ts, raw_body)
    return f"t={ts},v1={signature}"


def compute_signature(secret: str, timestamp: int, raw_body: str) -> str:
    base = f"{timestamp}.{raw_body}".encode("utf-8")
    return hmac.new(secret.encode("utf-8"), base, hashlib.sha256).hexdigest()


def _validate_signature(
    signature_header: str | None,
    root_path: Path,
    raw_body: str,
    *,
    ttl_seconds: int | None = None,
) -> None:
    secret = _resolve_webhook_secret(root_path)
    allow_unsigned = _allow_unsigned_webhooks()
    if not secret and not signature_header:
        if allow_unsigned:
            return
        raise WebhookSignatureError("signature_required", 401)
    if not secret:
        raise WebhookSignatureError("signature_secret_missing", 401)
    if not signature_header:
        raise WebhookSignatureError("signature_missing", 401)
    timestamp, signature = _parse_signature_header(signature_header)
    ttl = DEFAULT_SIGNATURE_TTL_SECONDS if ttl_seconds is None else ttl_seconds
    now = int(time.time())
    if ttl > 0 and abs(now - timestamp) > ttl:
        raise WebhookSignatureError("signature_expired", 410)
    expected = compute_signature(secret, timestamp, raw_body)
    if not hmac.compare_digest(signature, expected):
        raise WebhookSignatureError("invalid_signature", 401)


def _parse_signature_header(value: str) -> tuple[int, str]:
    header = value.strip()
    if header.lower().startswith("x-skillos-signature:"):
        header = header.split(":", 1)[1].strip()
    pairs: dict[str, str] = {}
    for part in header.split(","):
        if "=" not in part:
            continue
        key, val = part.split("=", 1)
        pairs[key.strip()] = val.strip()
    if "t" not in pairs or "v1" not in pairs:
        raise WebhookSignatureError("invalid_signature", 401)
    try:
        timestamp = int(pairs["t"])
    except ValueError as exc:
        raise WebhookSignatureError("invalid_signature", 401) from exc
    signature = pairs["v1"]
    if not signature:
        raise WebhookSignatureError("invalid_signature", 401)
    return timestamp, signature


def _resolve_webhook_secret(root_path: Path) -> str | None:
    resolver = SecretResolver(store=SecretsStore(default_secrets_path(root_path)))
    try:
        secret = resolver.resolve("secret", integration="webhook")
    except SecretResolutionError:
        return None
    return secret.value


def _allow_unsigned_webhooks() -> bool:
    raw = os.getenv("SKILLOS_WEBHOOK_ALLOW_UNSIGNED")
    if raw is None:
        return False
    return str(raw).strip().lower() in {"1", "true", "yes", "on"}


def _resolve_idempotency_key(
    payload: dict[str, object],
    override: str | None,
) -> str | None:
    if override:
        return str(override)
    key = payload.get("idempotency_key")
    if key is None:
        return None
    return str(key)


def _resolve_job_payload(
    payload: dict[str, object],
    raw: str,
    attachments: list[AttachmentReference],
) -> str:
    if attachments:
        return _build_attachment_payload(payload, attachments)
    if "payload" in payload:
        value = payload["payload"]
        if isinstance(value, str):
            return value
        return json.dumps(value, ensure_ascii=True)
    return raw


def _build_attachment_payload(
    payload: dict[str, object],
    attachments: list[AttachmentReference],
) -> str:
    base_payload = _resolve_base_payload(payload)
    payload_with_attachments = {
        "payload": base_payload,
        "attachments": [attachment.to_dict() for attachment in attachments],
    }
    return json.dumps(payload_with_attachments, ensure_ascii=True)


def _resolve_base_payload(payload: dict[str, object]) -> object:
    if "payload" in payload:
        return payload["payload"]
    return {
        key: value
        for key, value in payload.items()
        if key not in {"attachments", "idempotency_key"}
    }


def _check_idempotency(
    idempotency_key: str,
    trigger: WebhookTrigger,
    root_path: Path,
    *,
    ttl_seconds: int | None = None,
) -> IdempotencyDecision:
    store = idempotency_store_from_env(root_path)
    ttl = ttl_seconds if ttl_seconds is not None else idempotency_ttl_from_env()
    return store.check_and_record(
        "webhook",
        trigger.skill_id,
        idempotency_key,
        ttl_seconds=ttl,
    )



from __future__ import annotations

from dataclasses import dataclass
import os
import re
from typing import Callable, Iterable

from skillos.skills.models import SkillMetadata
from skillos.vector_search import (
    QdrantVectorSearch,
    VectorDocument,
    VectorSearchError,
    vector_search_config_from_env,
)

_TOKEN_RE = re.compile(r"[a-zA-Z0-9]+")

_KEYWORD_WEIGHTS: dict[str, int] = {
    "summarize": 3,
    "summary": 3,
    "summarise": 3,
    "convert": 3,
    "itinerary": 3,
    "quote": 3,
    "margin": 3,
    "exchange": 2,
    "rate": 2,
    "rates": 2,
    "logistics": 2,
    "shipping": 2,
    "delivery": 2,
    "hotel": 2,
    "hotels": 2,
    "flight": 2,
    "flights": 2,
    "tender": 2,
    "tenders": 2,
    "procurement": 2,
    "expenses": 2,
    "expense": 2,
    "web": 2,
    "search": 2,
    "document": 2,
    "requirements": 2,
    "certification": 2,
    "pricelist": 2,
    "price": 2,
    "prices": 2,
}


def to_public_id(skill_id: str) -> str:
    if "/" in skill_id:
        return skill_id.replace("/", ".", 1)
    return skill_id


def to_internal_id(skill_id: str) -> str:
    if "." in skill_id:
        return skill_id.replace(".", "/", 1)
    return skill_id


def _normalize_tokens(text: str) -> set[str]:
    tokens: set[str] = set()
    for match in _TOKEN_RE.finditer(text.lower()):
        token = match.group(0)
        tokens.add(token)
        if token.endswith("ies") and len(token) > 4:
            tokens.add(f"{token[:-3]}y")
        elif token.endswith("s") and len(token) > 3 and not token.endswith("ss"):
            tokens.add(token[:-1])
    return tokens


def _token_weight(token: str) -> int:
    return _KEYWORD_WEIGHTS.get(token, 1)


@dataclass(frozen=True)
class SkillProfile:
    internal_id: str
    public_id: str
    keywords: set[str]
    document: str
    tags: set[str]
    deprecated: bool

    @classmethod
    def from_metadata(cls, metadata: SkillMetadata) -> "SkillProfile":
        keywords: set[str] = set()
        document_parts: list[str] = []
        tags = _normalize_tag_filter(metadata.tags)
        for source in [
            metadata.id,
            metadata.name,
            metadata.description,
            *metadata.tags,
        ]:
            text = str(source)
            document_parts.append(text)
            keywords.update(_normalize_tokens(text))
        internal_id = metadata.id
        return cls(
            internal_id=internal_id,
            public_id=to_public_id(internal_id),
            keywords=keywords,
            document=" ".join(document_parts),
            tags=tags or set(),
            deprecated=metadata.deprecated,
        )


@dataclass
class SkillCandidate:
    skill_id: str
    score: float
    keyword_score: int = 0
    semantic_score: float = 0.0
    internal_id: str | None = None


@dataclass(frozen=True)
class RoutingConfig:
    mode: str = "hybrid"
    keyword_weight: float = 0.6
    vector_weight: float = 0.4
    vector_top_k: int = 8
    vector_min_score: float = 0.2
    include_deprecated: bool = False


def _env_float(name: str, default: float) -> float:
    raw = os.getenv(name)
    if raw is None:
        return default
    try:
        return float(raw)
    except ValueError:
        return default


def _env_int(name: str, default: int) -> int:
    raw = os.getenv(name)
    if raw is None:
        return default
    try:
        return int(raw)
    except ValueError:
        return default


def _env_bool(name: str, default: bool) -> bool:
    raw = os.getenv(name)
    if raw is None:
        return default
    return str(raw).strip().lower() in {"1", "true", "yes", "on"}


def routing_config_from_env() -> RoutingConfig:
    mode = os.getenv("SKILLOS_ROUTING_MODE", "hybrid").strip().lower()
    if mode not in {"keyword", "hybrid", "vector"}:
        mode = "keyword"
    vector_top_k = _env_int("SKILLOS_ROUTING_VECTOR_TOP_K", 8)
    vector_min_score = _env_float("SKILLOS_ROUTING_VECTOR_MIN_SCORE", 0.2)
    include_deprecated = _env_bool("SKILLOS_ROUTING_INCLUDE_DEPRECATED", False)
    if mode == "keyword":
        return RoutingConfig(
            mode=mode,
            keyword_weight=1.0,
            vector_weight=0.0,
            vector_top_k=vector_top_k,
            vector_min_score=vector_min_score,
            include_deprecated=include_deprecated,
        )
    if mode == "vector":
        return RoutingConfig(
            mode=mode,
            keyword_weight=0.0,
            vector_weight=1.0,
            vector_top_k=vector_top_k,
            vector_min_score=vector_min_score,
            include_deprecated=include_deprecated,
        )
    keyword_weight = _env_float("SKILLOS_ROUTING_KEYWORD_WEIGHT", 0.6)
    vector_weight = _env_float("SKILLOS_ROUTING_VECTOR_WEIGHT", 0.4)
    return RoutingConfig(
        mode=mode,
        keyword_weight=keyword_weight,
        vector_weight=vector_weight,
        vector_top_k=vector_top_k,
        vector_min_score=vector_min_score,
        include_deprecated=include_deprecated,
    )


@dataclass(frozen=True)
class RoutingResult:
    status: str
    skill_id: str | None
    internal_skill_id: str | None
    confidence: float
    candidates: list[SkillCandidate]
    alternatives: list[str]


class SkillRouter:
    def __init__(
        self,
        skills: Iterable[SkillMetadata],
        low_confidence_threshold: float = 0.25,
        margin_threshold: float = 0.05,
        confidence_provider: Callable[[str], float] | None = None,
        vector_search: QdrantVectorSearch | None = None,
        routing_config: RoutingConfig | None = None,
    ) -> None:
        profiles = [SkillProfile.from_metadata(metadata) for metadata in skills]
        self._profiles: dict[str, SkillProfile] = {
            profile.public_id: profile for profile in profiles
        }
        self._low_confidence_threshold = low_confidence_threshold
        self._margin_threshold = margin_threshold
        self._confidence_provider = confidence_provider
        self._vector_search = vector_search
        self._routing_config = routing_config or RoutingConfig()
        if self._vector_search:
            self._index_vector_search()

    def _adjust_score(self, score: float, skill_id: str) -> float:
        if not self._confidence_provider:
            return score
        confidence = self._confidence_provider(skill_id)
        multiplier = 1 + (confidence - 0.5) * 3
        if multiplier < 0.1:
            multiplier = 0.1
        elif multiplier > 2.0:
            multiplier = 2.0
        return score * multiplier

    def rank_candidates(
        self, query: str, candidates: Iterable[SkillCandidate]
    ) -> list[SkillCandidate]:
        query_tokens = _normalize_tokens(query)
        ranked: list[SkillCandidate] = []
        for candidate in candidates:
            public_id = to_public_id(candidate.skill_id)
            profile = self._profiles.get(public_id)
            keyword_score = candidate.keyword_score
            internal_id = candidate.internal_id
            if profile:
                if keyword_score == 0:
                    keyword_score = self._keyword_score(query_tokens, profile.keywords)
                if internal_id is None:
                    internal_id = profile.internal_id
            adjusted_score = self._adjust_score(candidate.score, public_id)
            ranked.append(
                SkillCandidate(
                    skill_id=public_id,
                    score=adjusted_score,
                    keyword_score=keyword_score,
                    semantic_score=candidate.semantic_score,
                    internal_id=internal_id,
                )
            )
        return sorted(
            ranked,
            key=lambda item: (-item.score, -item.keyword_score, item.skill_id),
        )

    def route(
        self,
        query: str,
        limit: int = 5,
        tags: Iterable[str] | None = None,
    ) -> RoutingResult:
        query_tokens = _normalize_tokens(query)
        keyword_candidates = self._keyword_candidates(query_tokens, tags)
        if not self._vector_search or self._routing_config.mode == "keyword":
            ranked = self.rank_candidates(query, keyword_candidates.values())[:limit]
            return self._build_result(ranked)

        semantic_scores = self._semantic_scores(query, tags)
        if not semantic_scores:
            ranked = self.rank_candidates(query, keyword_candidates.values())[:limit]
            return self._build_result(ranked)

        keyword_weight, vector_weight = self._normalized_weights(
            self._routing_config.keyword_weight,
            self._routing_config.vector_weight,
        )
        if vector_weight == 0.0:
            ranked = self.rank_candidates(query, keyword_candidates.values())[:limit]
            return self._build_result(ranked)

        candidates: list[SkillCandidate] = []
        for skill_id in set(keyword_candidates) | set(semantic_scores):
            keyword_candidate = keyword_candidates.get(skill_id)
            keyword_score = keyword_candidate.keyword_score if keyword_candidate else 0
            keyword_norm = keyword_candidate.score if keyword_candidate else 0.0
            semantic_score = semantic_scores.get(skill_id, 0.0)
            combined_score = (
                keyword_norm * keyword_weight + semantic_score * vector_weight
            )
            if combined_score <= 0.0:
                continue
            internal_id = (
                keyword_candidate.internal_id
                if keyword_candidate and keyword_candidate.internal_id
                else self._profiles[skill_id].internal_id
            )
            candidates.append(
                SkillCandidate(
                    skill_id=skill_id,
                    score=combined_score,
                    keyword_score=keyword_score,
                    semantic_score=semantic_score,
                    internal_id=internal_id,
                )
            )

        ranked = self.rank_candidates(query, candidates)[:limit]
        return self._build_result(ranked)

    def _build_result(self, ranked: list[SkillCandidate]) -> RoutingResult:
        if not ranked:
            return RoutingResult(
                status="no_skill_found",
                skill_id=None,
                internal_skill_id=None,
                confidence=0.0,
                candidates=[],
                alternatives=[],
            )

        top = ranked[0]
        confidence = top.score
        alternatives = [candidate.skill_id for candidate in ranked[1:3]]
        status = "selected"
        if confidence < self._low_confidence_threshold:
            status = "low_confidence"
        elif len(ranked) > 1 and (top.score - ranked[1].score) < self._margin_threshold:
            status = "low_confidence"

        return RoutingResult(
            status=status,
            skill_id=top.skill_id,
            internal_skill_id=top.internal_id,
            confidence=confidence,
            candidates=ranked,
            alternatives=alternatives,
        )

    @staticmethod
    def _keyword_score(query_tokens: set[str], keywords: set[str]) -> int:
        return sum(
            _token_weight(token) for token in query_tokens if token in keywords
        )

    def _keyword_candidates(
        self,
        query_tokens: set[str],
        tags: Iterable[str] | None = None,
    ) -> dict[str, SkillCandidate]:
        total_weight = sum(_token_weight(token) for token in query_tokens) or 1
        candidates: dict[str, SkillCandidate] = {}
        tag_filter = _normalize_tag_filter(tags)
        for profile in self._profiles.values():
            if not self._routing_config.include_deprecated and profile.deprecated:
                continue
            if tag_filter and not _matches_tags(profile.tags, tag_filter):
                continue
            keyword_score = self._keyword_score(query_tokens, profile.keywords)
            if keyword_score == 0:
                continue
            score = keyword_score / total_weight
            candidates[profile.public_id] = SkillCandidate(
                skill_id=profile.public_id,
                score=score,
                keyword_score=keyword_score,
                internal_id=profile.internal_id,
            )
        return candidates

    def _semantic_scores(
        self, query: str, tags: Iterable[str] | None = None
    ) -> dict[str, float]:
        if not self._vector_search:
            return {}
        try:
            matches = self._vector_search.search(
                query,
                limit=self._routing_config.vector_top_k,
                min_score=self._routing_config.vector_min_score,
            )
        except VectorSearchError:
            return {}
        scores: dict[str, float] = {}
        tag_filter = _normalize_tag_filter(tags)
        for match in matches:
            public_id = to_public_id(match.doc_id)
            if public_id not in self._profiles:
                continue
            profile = self._profiles[public_id]
            if (
                not self._routing_config.include_deprecated
                and profile.deprecated
            ):
                continue
            if tag_filter and not _matches_tags(
                profile.tags,
                tag_filter,
            ):
                continue
            scores[public_id] = match.score
        return scores

    def _normalized_weights(self, keyword_weight: float, vector_weight: float) -> tuple[float, float]:
        keyword_weight = max(0.0, keyword_weight)
        vector_weight = max(0.0, vector_weight)
        total = keyword_weight + vector_weight
        if total == 0.0:
            return 1.0, 0.0
        return keyword_weight / total, vector_weight / total

    def _index_vector_search(self) -> None:
        if not self._vector_search:
            return
        documents = [
            VectorDocument(
                doc_id=profile.public_id,
                text=profile.document,
                payload={"skill_id": profile.public_id},
            )
            for profile in self._profiles.values()
        ]
        try:
            self._vector_search.index(documents)
        except VectorSearchError:
            self._vector_search = None


def build_router_from_env(
    skills: Iterable[SkillMetadata],
    *,
    confidence_provider: Callable[[str], float] | None = None,
) -> SkillRouter:
    routing_config = routing_config_from_env()
    vector_search = None
    if routing_config.mode != "keyword":
        vector_config = vector_search_config_from_env()
        if vector_config:
            vector_search = QdrantVectorSearch(vector_config)
    return SkillRouter(
        skills,
        confidence_provider=confidence_provider,
        vector_search=vector_search,
        routing_config=routing_config,
    )


def _normalize_tag_filter(tags: Iterable[str] | None) -> set[str] | None:
    if not tags:
        return None
    normalized = {str(tag).strip().lower() for tag in tags if str(tag).strip()}
    return normalized or None


def _matches_tags(skill_tags: set[str], tag_filter: set[str]) -> bool:
    return bool(skill_tags.intersection(tag_filter))



from __future__ import annotations
from pathlib import Path
import click

from skillos.skills.paths import default_skills_root
from skillos.debugging import DebugTraceConfig, DebugTrace, StepController, render_trace
from skillos.orchestrator import Orchestrator
from skillos.cli.utils import (
    _apply_auth_context,
    _emit_warnings,
    _blocked_message
)


@click.command("run")
@click.argument("query")
@click.option(
    "--root",
    "root_path",
    type=click.Path(file_okay=False, path_type=Path),
    default=default_skills_root(),
    show_default=True,
)
@click.option(
    "--log-path",
    "log_path",
    type=click.Path(dir_okay=False, path_type=Path),
    default=None,
)
@click.option("--execute/--no-execute", default=False, show_default=True)
@click.option("--dry-run/--no-dry-run", default=False, show_default=True)
@click.option(
    "--approval",
    type=click.Choice(["approved", "denied"], case_sensitive=False),
    default=None,
)
@click.option("--approval-token", default=None)
@click.option("--role", default=None)
@click.option("--tag", "tags", multiple=True)
@click.option(
    "--mode",
    type=click.Choice(["single", "pipeline", "parallel", "auto"], case_sensitive=False),
    default="single",
    show_default=True,
)
@click.option(
    "--plan-path",
    "plan_path",
    type=click.Path(dir_okay=False, path_type=Path),
    default=None,
)
@click.option("--debug/--no-debug", default=False, show_default=True)
@click.option("--profile/--no-profile", default=False, show_default=True)
@click.option(
    "--trace/--no-trace",
    "show_trace",
    default=False,
    show_default=True,
)
@click.option(
    "--step-through/--no-step-through",
    "step_through",
    default=False,
    show_default=True,
)
@click.option("--jwt", "jwt_token", default=None)
def run_query(
    query: str,
    root_path: Path,
    log_path: Path | None,
    execute: bool,
    dry_run: bool,
    approval: str | None,
    approval_token: str | None,
    role: str | None,
    tags: tuple[str, ...],
    mode: str,
    plan_path: Path | None,
    debug: bool,
    profile: bool,
    show_trace: bool,
    step_through: bool,
    jwt_token: str | None,
) -> DebugTrace | None:
    """Route a query to the best skill."""
    
    trace_config = DebugTraceConfig(
        capture_inputs=debug or step_through,
        capture_outputs=debug or step_through,
        capture_timing=profile,
    )
    step_controller = None
    if step_through:
        step_controller = StepController(
            show_inputs=debug or step_through,
            show_outputs=debug or step_through,
            show_timing=profile,
        )
    debug_trace = (
        DebugTrace(trace_config, step_controller=step_controller)
        if (debug or profile or show_trace or step_through)
        else None
    )
    
    try:
        root_path, role, attributes = _apply_auth_context(
            root_path, role, jwt_token
        )
        orchestrator = Orchestrator(root_path, log_path)
        result = orchestrator.run_query(
            query=query,
            execute=execute,
            dry_run=dry_run,
            approval=approval,
            approval_token=approval_token,
            role=role,
            attributes=attributes,
            tags=list(tags) if tags else None,
            mode=mode,
            plan_path=plan_path,
            debug_trace=debug_trace,
        )
        
        status = result.get("status")
        _emit_warnings(result.get("warnings"))
        
        if status == "no_skill_found":
            click.echo("no_skill_found")
            return debug_trace
            
        if status == "low_confidence":
            click.echo(f"low_confidence: {result['skill_id']}")
            alternatives = ", ".join(result.get("alternatives", [])) or "none"
            click.echo(f"alternatives: {alternatives}")
            return debug_trace
            
        if status == "blocked":
            reason = result.get("reason") or result.get("policy_id") or "blocked"
            click.echo(f"blocked: {_blocked_message(reason)}")
            return debug_trace
            
        # Success or Dry Run or Routed
        click.echo(result.get("skill_id"))

        if status == "dry_run":
            if result.get("preview"):
                preview = result["preview"]
                click.echo(f"plan_id: {preview.plan_id}")
                click.echo("affected_entities:")
                for entity in preview.affected_entities:
                    click.echo(f"- {entity}")
            if result.get("plan_path"):
                 click.echo(f"plan_written: {result['plan_path']}")
        
        elif status == "success":
            if result.get("plan_id"):
                 click.echo(f"plan_id: {result['plan_id']}")
            click.echo(result.get("output"))
        elif status == "error":
            raise click.ClickException(result.get("reason") or "execution_failed")
            
    except click.ClickException:
        raise
    except Exception as e:
        if debug:
            raise
        click.echo(f"Error: {str(e)}", err=True)
    finally:
        if show_trace and debug_trace:
            click.echo(render_trace(debug_trace))
            
    return debug_trace



from __future__ import annotations

from dataclasses import dataclass
import hashlib
import os
import re
from pathlib import Path

from skillos.tenancy import tenant_id_from_path, tenant_id_from_env


_SUPPORTED_BACKENDS = {"file", "postgres"}
_SCHEMA_RE = re.compile(r"^[A-Za-z_][A-Za-z0-9_]*$")


@dataclass(frozen=True)
class StorageBackendConfig:
    backend: str
    postgres_dsn: str | None
    postgres_schema: str


def storage_backend_from_env() -> StorageBackendConfig:
    backend = os.getenv("SKILLOS_STORAGE_BACKEND", "file").strip().lower()
    if backend not in _SUPPORTED_BACKENDS:
        backend = "file"
    dsn = os.getenv("SKILLOS_POSTGRES_DSN") or os.getenv("DATABASE_URL")
    schema = (os.getenv("SKILLOS_POSTGRES_SCHEMA") or "skillos").strip()
    if not _SCHEMA_RE.match(schema):
        schema = "skillos"
    return StorageBackendConfig(
        backend=backend,
        postgres_dsn=dsn,
        postgres_schema=schema,
    )


def require_postgres_dsn(config: StorageBackendConfig, *, context: str) -> str:
    if not config.postgres_dsn:
        raise ValueError(f"{context}_postgres_dsn_missing")
    return config.postgres_dsn


def resolve_tenant_id(root: Path) -> str:
    tenant_id = tenant_id_from_path(root) or tenant_id_from_env()
    if tenant_id:
        return tenant_id
    test_id = os.getenv("PYTEST_CURRENT_TEST")
    if test_id:
        digest = hashlib.blake2b(test_id.encode("utf-8"), digest_size=6).hexdigest()
        return f"test-{digest}"
    return "default"


def pg_connect(dsn: str):
    import psycopg
    from psycopg.rows import dict_row

    return psycopg.connect(dsn, row_factory=dict_row)



from __future__ import annotations

from dataclasses import dataclass
import base64
from pathlib import Path
from uuid import uuid4

from skillos.tenancy import resolve_tenant_root

class AttachmentError(ValueError):
    pass


@dataclass(frozen=True)
class AttachmentInput:
    filename: str
    content_type: str
    data: bytes


_DEFAULT_MAX_SIZE_BYTES = 10 * 1024 * 1024  # 10 MB
_ALLOWED_CONTENT_TYPES = {
    "application/json",
    "application/pdf",
    "text/plain",
    "text/csv",
    "text/markdown",
    "text/html",  # Careful with this one, but often needed
    "image/jpeg",
    "image/png",
    "image/gif",
    "image/webp",
}


@dataclass(frozen=True)
class AttachmentReference:
    attachment_id: str
    filename: str
    content_type: str
    size_bytes: int
    reference: str

    def to_dict(self) -> dict[str, object]:
        return {
            "id": self.attachment_id,
            "filename": self.filename,
            "content_type": self.content_type,
            "size_bytes": self.size_bytes,
            "reference": self.reference,
        }


def default_attachments_path(root: Path) -> Path:
    root_path = resolve_tenant_root(root)
    return root_path / "attachments"


def ingest_attachments(
    raw: object,
    root: Path,
    *,
    request_id: str | None = None,
) -> list[AttachmentReference]:
    attachments = _parse_attachments(raw)
    if not attachments:
        return []

    root_path = Path(root)
    attachments_root = default_attachments_path(root_path)
    batch_id = request_id or uuid4().hex
    batch_dir = attachments_root / batch_id
    batch_dir.mkdir(parents=True, exist_ok=True)

    references: list[AttachmentReference] = []
    for attachment in attachments:
        attachment_id = uuid4().hex
        stored_name = f"{attachment_id}_{attachment.filename}"
        stored_path = batch_dir / stored_name
        stored_path.write_bytes(attachment.data)
        reference = (Path("attachments") / batch_id / stored_name).as_posix()
        references.append(
            AttachmentReference(
                attachment_id=attachment_id,
                filename=attachment.filename,
                content_type=attachment.content_type,
                size_bytes=len(attachment.data),
                reference=reference,
            )
        )
    return references


def _parse_attachments(raw: object) -> list[AttachmentInput]:
    if raw is None:
        return []
    if not isinstance(raw, list):
        raise AttachmentError("invalid_attachments")

    attachments: list[AttachmentInput] = []
    for item in raw:
        attachments.append(_parse_attachment(item))
    return attachments


def _parse_attachment(item: object) -> AttachmentInput:
    if not isinstance(item, dict):
        raise AttachmentError("invalid_attachment")

    filename = _normalize_filename(item.get("filename") or item.get("name"))
    if not filename:
        raise AttachmentError("invalid_attachment_name")

    content_type = _normalize_content_type(
        item.get("content_type") or item.get("type")
    )
    if not content_type:
        raise AttachmentError("invalid_attachment_type")

    data = item.get("data")
    if not isinstance(data, (str, bytes)):
        raise AttachmentError("invalid_attachment_data")
    try:
        payload = data.encode("utf-8") if isinstance(data, str) else data
        decoded = base64.b64decode(payload, validate=True)
    except ValueError as exc:
        raise AttachmentError("invalid_attachment_data") from exc

    if not decoded:
        raise AttachmentError("invalid_attachment_data")

    _validate_attachment_limits(content_type, len(decoded))

    return AttachmentInput(
        filename=filename,
        content_type=content_type,
        data=decoded,
    )


def _validate_attachment_limits(
    content_type: str,
    data_len: int,
) -> None:
    if content_type not in _ALLOWED_CONTENT_TYPES:
        # Strict allowlist
        raise AttachmentError(f"invalid_attachment_type: {content_type}")

    max_size = _env_int("SKILLOS_ATTACHMENT_MAX_SIZE_BYTES", _DEFAULT_MAX_SIZE_BYTES)
    if data_len > max_size:
        raise AttachmentError("attachment_too_large")


def _env_int(name: str, default: int) -> int:
    import os
    raw = os.getenv(name)
    if raw is None:
        return default
    try:
        return int(str(raw).strip())
    except ValueError:
        return default


def _normalize_filename(value: object) -> str:
    if value is None:
        return ""
    name = str(value).strip()
    if not name:
        return ""
    return Path(name).name


def _normalize_content_type(value: object) -> str:
    if value is None:
        return ""
    content_type = str(value).strip().lower()
    return content_type



from __future__ import annotations

from dataclasses import dataclass
import re

HIGH_RISK_THRESHOLD = 11
CRITICAL_RISK_THRESHOLD = 16
MEDIUM_RISK_THRESHOLD = 6
_HIGH_RISK_SCORE = 12
_LOW_RISK_SCORE = 2

_HIGH_RISK_PATTERNS: dict[str, re.Pattern[str]] = {
    "delete": re.compile(r"\bdelete\b"),
    "remove": re.compile(r"\bremove\b"),
    "drop": re.compile(r"\bdrop\b"),
    "bulk_update": re.compile(r"\bbulk[-\s]+update\b"),
    "update_all": re.compile(r"\bupdate\s+all\b"),
    "mass_update": re.compile(r"\bmass\s+update\b"),
}


@dataclass(frozen=True)
class RiskAssessment:
    score: int
    level: str
    reasons: list[str]


class RiskScorer:
    def __init__(
        self,
        high_risk_threshold: int = HIGH_RISK_THRESHOLD,
        critical_risk_threshold: int = CRITICAL_RISK_THRESHOLD,
    ) -> None:
        self._high_risk_threshold = high_risk_threshold
        self._critical_risk_threshold = critical_risk_threshold

    def assess(self, text: str, skill_id: str | None = None) -> RiskAssessment:
        combined = " ".join(part for part in [text, skill_id] if part)
        normalized = combined.lower()
        reasons = [
            name for name, pattern in _HIGH_RISK_PATTERNS.items() if pattern.search(normalized)
        ]
        score = _HIGH_RISK_SCORE if reasons else _LOW_RISK_SCORE
        level = self._level_for(score)
        return RiskAssessment(score=score, level=level, reasons=reasons)

    def _level_for(self, score: int) -> str:
        if score >= self._critical_risk_threshold:
            return "critical"
        if score >= self._high_risk_threshold:
            return "high"
        if score >= MEDIUM_RISK_THRESHOLD:
            return "medium"
        return "low"



from __future__ import annotations

from dataclasses import dataclass
import hashlib
import json
from pathlib import Path

from skillos.storage import atomic_write_text

@dataclass(frozen=True)
class ExecutionPlan:
    skill_id: str
    internal_skill_id: str
    payload: str

    @property
    def plan_id(self) -> str:
        payload = json.dumps(self.to_dict(), sort_keys=True, ensure_ascii=True)
        return hashlib.sha256(payload.encode("utf-8")).hexdigest()

    def to_dict(self) -> dict[str, str]:
        return {
            "skill_id": self.skill_id,
            "internal_skill_id": self.internal_skill_id,
            "payload": self.payload,
        }

    @classmethod
    def from_dict(cls, payload: dict[str, str]) -> "ExecutionPlan":
        return cls(
            skill_id=payload["skill_id"],
            internal_skill_id=payload["internal_skill_id"],
            payload=payload["payload"],
        )


def build_execution_plan(
    skill_id: str, internal_skill_id: str, payload: str
) -> ExecutionPlan:
    return ExecutionPlan(
        skill_id=skill_id,
        internal_skill_id=internal_skill_id,
        payload=payload,
    )


def save_execution_plan(plan: ExecutionPlan, path: Path) -> None:
    path = Path(path)
    atomic_write_text(
        path,
        json.dumps(plan.to_dict(), ensure_ascii=True),
        encoding="utf-8",
    )


def load_execution_plan(path: Path) -> ExecutionPlan:
    payload = json.loads(Path(path).read_text(encoding="utf-8"))
    return ExecutionPlan.from_dict(payload)
